# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Constants used for interaction with defined entity framework."""
from enum import Enum
from enum import unique

import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501

CLUSTER_ACL_LIST_FIELDS = [shared_constants.AccessControlKey.ACCESS_LEVEL_ID,
                           shared_constants.AccessControlKey.MEMBER_ID,
                           shared_constants.AccessControlKey.USERNAME]

# ACL Path
ACTION_CONTROL_ACCESS_PATH = '/action/controlAccess/'
DEF_INTERFACE_ID_PREFIX = 'urn:vcloud:interface'
DEF_NATIVE_ENTITY_TYPE_NSS = 'nativeCluster'
DEF_ENTITY_TYPE_ID_PREFIX = 'urn:vcloud:type'
DEF_API_MIN_VERSION = 35.0
DEF_SCHEMA_DIRECTORY = 'cse_def_schema'
DEF_ERROR_MESSAGE_KEY = 'message'
DEF_RESOLVED_STATE = 'RESOLVED'

PAYLOAD_VERSION_PREFIX = 'cse.vmware.com/'
PAYLOAD_VERSION_2_0 = PAYLOAD_VERSION_PREFIX + 'v2.0'


@unique
class Vendor(str, Enum):
    CSE = 'cse'
    VMWARE = 'vmware'


@unique
class Nss(str, Enum):
    KUBERNETES = 'k8s'
    NATIVE_CLUSTER = 'nativeCluster'
    TKG = 'tkgcluster'


@unique
class SchemaFile(str, Enum):
    SCHEMA_1_0_0 = 'schema_1_0_0.json'
    SCHEMA_2_0_0 = 'schema_2_0_0.json'


DEF_NATIVE_ENTITY_TYPE_RIGHT_BUNDLE = \
    f'{Vendor.CSE.value}:{Nss.NATIVE_CLUSTER.value} Entitlement'


# Defines the RDE version CSE server uses at runtime.
# RDE version used by a given CSE version can differ based on the VCD
# environment it is configured with.
#
# Examples on how to updated this Enum:
#
# 1. An existing entry must be updated on a minor or patch version
# increment of RDE, example: RDE_2_X = '2.1.0'.
# 2. A New Entry must be added on each major version increment of
# RDE, example RDE_3_X = '3.0.0'
#
# Any updates to this enum may also require updating common_models.MAP_RDE_VERSION_TO_ITS_METADATA  # noqa: E501
@unique
class RuntimeRDEVersion(str, Enum):
    RDE_1_X = '1.0.0'
    RDE_2_X = '2.0.0'


@unique
class RDEVersion(str, Enum):
    RDE_1_0_0 = '1.0.0'
    RDE_2_0_0 = '2.0.0'


@unique
class RDEMetadataKey(str, Enum):
    ENTITY_TYPE = 'entity_type'
    INTERFACES = 'interfaces'
    INTERFACE_TO_BEHAVIORS_MAP = 'map_from_interface_id_to_behaviors'
    ENTITY_TYPE_TO_OVERRIDABLE_BEHAVIORS_MAP = 'map_from_entity_type_id_to_behaviors_to_override'  # noqa: E501
    BEHAVIOR_TO_ACL_MAP = 'map_from_behavior_id_to_behavior_acls'


# Below map essentially answers this question:
# "When CSE is configured with VCD of a given api-version, what is the
# compatible RDE-version CSE-server must use on that VCD?"
# Based on the VCD-version CSE server is configured with,
# CSE server dynamically determines the RDE-version-to-use at runtime.
#
# Key: represents the VCD api-version CSE is configured with.
# Value: represents the RDE-version CSE-server uses for that VCD environment.
#
# This map must be carefully updated for every major/minor/patch version
# increments of RDE for each official CSE release. Below are few examples on
#  how this map could be updated.
# NOTE: The RDE version used by the VCD API version can be different in other
#   CSE releases.

# Examples:
# Mapping for CSE 3.1:
# MAP_VCD_API_VERSION_TO_RUNTIME_RDE_VERSION = {
#     35.0: 1.0.0,
#     36.0: 2.0.0
# }
#
# If CSE 3.2 introduces Minor version bump in RDE (i.e 2.1) and is released
#   alongside vCD 10.4 (API Version 37), mapping would become -
# MAP_VCD_API_VERSION_TO_RUNTIME_RDE_VERSION = {
# 35.0: 1.0.0,
# 36.0: 2.1.0 (Note the RDE version change),
# 37.0: 2.1.0 (Newly introduced RDE),
# }
#
# If CSE 3.2 introduces Major version bump in RDE (i.e 3.0) and is released
#   alongside vCD 10.4 (API version 37), mapping would be -
# MAP_VCD_API_VERSION_TO_RUNTIME_RDE_VERSION = {
# 35.0: 1.0.0,
# 36.0: 2.0.0,
# 37.0: 3.0.0 (Newly introduced RDE)
# }
MAP_VCD_API_VERSION_TO_RUNTIME_RDE_VERSION = {
    '35.0': RuntimeRDEVersion.RDE_1_X.value,
    '36.0': RuntimeRDEVersion.RDE_2_X.value
}

# Below map answers the question - "what is the RDE version introduced by CSE
# at a given VCD api version?"
#
# Key: represents the VCD api_version.
# Value: RDE version that is introduced at the specified api version.
#
# For every new RDE version, a new entry must be added to this map.
# Example: Say, at 37 api version, RDE version 2.1 is introduced.
# The map will change to the below.
# MAP_VCD_API_VERSION_TO_RDE_VERSION = {
#     35.0: RDEVersion.RDE_1_0_0,
#     36.0: RDEVersion.RDE_2_0_0,
#     37.0: RDEVersion.RDE_3_0_0
# }
MAP_VCD_API_VERSION_TO_RDE_VERSION = {
    '35.0': RDEVersion.RDE_1_0_0.value,
    '36.0': RDEVersion.RDE_2_0_0.value
}

MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION = {
    PAYLOAD_VERSION_2_0: RDEVersion.RDE_2_0_0.value
}


class ClusterEntityFilterKey1X(Enum):
    """Keys to filter cluster entities in CSE (or) vCD.

    NOTE: Use ClusterEntityFilterKeyV1 if RDE version is <= 1.0.0
    else, use ClusterEntityFilterKey

    Below Keys are commonly used filters. An entity can be filtered by any of
    its properties.

    Usage examples:
    ..api/cse/internal/clusters?entity.kind=native
    ..api/cse/internal/clusters?entity.metadata.org_name=org1
    ..cloudapi/1.0.0/entities?filter=entity.metadata.org_name==org1
    """

    # TODO(DEF) CLI can leverage this enum for the filter implementation.
    CLUSTER_NAME = 'name'
    ORG_NAME = 'entity.metadata.org_name'
    OVDC_NAME = 'entity.metadata.ovdc_name'
    KIND = 'entity.kind'
    K8_DISTRIBUTION = 'entity.spec.k8_distribution.template_name'
    STATE = 'state'
    PHASE = 'entity.status.phase'


class ClusterEntityFilterKey2X(Enum):
    """Keys to filter cluster entities in CSE (or) vCD.

    Below Keys are commonly used filters. An entity can be filtered by any of
    its properties.

    Usage examples:
    ..api/cse/internal/clusters?entity.kind=native
    ..api/cse/internal/clusters?entity.metadata.orgName=org1
    ..cloudapi/1.0.0/entities?filter=entity.metadata.orgName==org1
    """

    # TODO(DEF) CLI can leverage this enum for the filter implementation.
    CLUSTER_NAME = 'name'
    ORG_NAME = 'entity.metadata.orgName'
    OVDC_NAME = 'entity.metadata.virtualDataCenterName'
    KIND = 'entity.kind'
    K8_DISTRIBUTION = 'entity.spec.distribution.templateName'
    STATE = 'state'
    PHASE = 'entity.status.phase'


class RDEFilterKey(Enum):
    """Keys to filter RDE."""

    NAME = 'name'
    STATE = 'state'


class PayloadKey(str, Enum):
    PAYLOAD_VERSION = 'apiVersion'
    PAYLOAD_VERSION_RDE_1_0 = 'api_version'
    UNKNOWN = 'unknown'



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
import hashlib

from pyvcloud.vcd.api_extension import APIExtension
from pyvcloud.vcd.client import BasicLoginCredentials
from pyvcloud.vcd.client import Client
import requests

from container_service_extension.common.constants.server_constants import CSE_SERVICE_NAME  # noqa: E501
from container_service_extension.common.constants.server_constants import CSE_SERVICE_NAMESPACE  # noqa: E501
from container_service_extension.common.constants.server_constants import MQTT_EXTENSION_VENDOR  # noqa: E501
from container_service_extension.common.constants.server_constants import MQTT_EXTENSION_VERSION  # noqa: E501
from container_service_extension.common.constants.server_constants import MQTTExtKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import SYSTEM_ORG_NAME  # noqa: E501
from container_service_extension.common.utils.core_utils import NullPrinter
from container_service_extension.common.utils.server_utils import should_use_mqtt_protocol  # noqa: E501
from container_service_extension.lib.telemetry.constants import COLLECTOR_ID
from container_service_extension.lib.telemetry.constants import VAC_URL
from container_service_extension.logging.logger import NULL_LOGGER
from container_service_extension.mqi.mqtt_extension_manager import \
    MQTTExtensionManager


CEIP_HEADER_NAME = "x-vmware-vcloud-ceip-id"


def uuid_hash(uuid):
    """Return SHA1 hash as hex digest of an uuid.

    Requirement from VAC team : You should apply SHA1 hashing over the data as
    a text. More specifically you should have the text data as a UTF8 encoded
    string then convert the string to byte array and digest it into a SHA1
    hash. The resulting SHA1 hash must be converted to text by displaying each
    byte of the hashcode as a HEX char (first byte displayed leftmost in the
    output). The hash must be lowercase.

    No checks are made to determine if the input uuid is valid or not. Dashes
    in the uuid are ignored while computing the hash.

    :param str uuid: uuid to be hashed

    :returns: SHA1 hash as hex digest of the provided uuid.
    """
    uuid_no_dash = uuid.replace('-', '')
    m = hashlib.sha1()
    m.update(bytes(uuid_no_dash, 'utf-8'))
    return m.hexdigest()


def get_vcd_ceip_id(vcd_host, verify_ssl=True, logger_debug=NULL_LOGGER):
    """."""
    response = None
    try:
        if not verify_ssl:
            requests.packages.urllib3.disable_warnings()
        uri = f"https://{vcd_host}"
        response = requests.get(uri, verify=verify_ssl)
        return response.headers.get(CEIP_HEADER_NAME)
    except Exception as err:
        logger_debug.error(f"Unable to get vCD CEIP id : {str(err)}")
    finally:
        if response:
            response.close()


def get_telemetry_instance_id(config_dict, logger_debug=NULL_LOGGER,
                              msg_update_callback=NullPrinter()):
    """Get CSE AMQP or MQTT extension id which is used as instance id.

    Any exception is logged as error. No exception is leaked out
    of this method and does not affect the server startup.

    :param dict config_dict: CSE configuration
    :param logging.logger logger_debug: logger instance to log any error
    in retrieving CSE extension id.
    :param utils.ConsoleMessagePrinter msg_update_callback: Callback object.

    :return instance id to use for sending data to Vmware telemetry server

    :rtype str (unless no instance id found)
    """
    vcd = config_dict['vcd']
    client = None
    try:
        client = Client(vcd['host'], verify_ssl_certs=vcd['verify'])
        client.set_credentials(BasicLoginCredentials(
            vcd['username'], SYSTEM_ORG_NAME, vcd['password']))
        if should_use_mqtt_protocol(config_dict):
            # Get MQTT extension uuid
            mqtt_ext_manager = MQTTExtensionManager(client)
            ext_info = mqtt_ext_manager.get_extension_info(
                ext_name=CSE_SERVICE_NAME,
                ext_version=MQTT_EXTENSION_VERSION,
                ext_vendor=MQTT_EXTENSION_VENDOR)
            if not ext_info:
                logger_debug.debug("Failed to retrieve telemetry instance id")
                return None
            logger_debug.debug("Retrieved telemetry instance id")
            return mqtt_ext_manager.get_extension_uuid(
                ext_info[MQTTExtKey.EXT_URN_ID])
        else:
            # Get AMQP extension id
            ext = APIExtension(client)
            cse_info = ext.get_extension_info(CSE_SERVICE_NAME,
                                              namespace=CSE_SERVICE_NAMESPACE)
            logger_debug.debug("Retrieved telemetry instance id")
            return cse_info.get('id')
    except Exception as err:
        msg = f"Cannot retrieve telemetry instance id:{err}"
        msg_update_callback.general(msg)
        logger_debug.error(msg, exc_info=True)
    finally:
        if client is not None:
            client.logout()


def store_telemetry_settings(config_dict):
    """Populate telemetry instance id, url and collector id in config.

    :param dict config_dict: CSE configuration
    """
    if 'vac_url' not in config_dict['service']['telemetry']:
        config_dict['service']['telemetry']['vac_url'] = VAC_URL

    config_dict['service']['telemetry']['collector_id'] = COLLECTOR_ID

    vcd_ceip_id = None
    instance_id = None
    if config_dict['service']['telemetry']['enable']:
        vcd_ceip_id = get_vcd_ceip_id(config_dict['vcd']['host'],
                                      verify_ssl=config_dict['vcd']['verify'])
        instance_id = get_telemetry_instance_id(config_dict)
    config_dict['service']['telemetry']['vcd_ceip_id'] = vcd_ceip_id
    config_dict['service']['telemetry']['instance_id'] = instance_id



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from container_service_extension.lib.pksclient.api.cluster_api import ClusterApi  # noqa: E501
from container_service_extension.lib.pksclient.api.plans_api import PlansApi
from container_service_extension.lib.pksclient.api.profile_api import ProfileApi  # noqa: E501
from container_service_extension.lib.pksclient.api.quota_api import QuotaApi
from container_service_extension.lib.pksclient.api.task_api import TaskApi
from container_service_extension.lib.pksclient.api.upgradable_api import UpgradableApi # noqa: E501
from container_service_extension.lib.pksclient.api.usage_api import UsageApi
from container_service_extension.lib.pksclient.api.users_api import UsersApi


# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import abc


class AbstractNativeEntity(abc.ABC):

    @classmethod
    @abc.abstractmethod
    def from_native_entity(cls, native_entity):
        pass

    @classmethod
    @abc.abstractmethod
    def from_cluster_data(cls, cluster: dict, kind: str, **kwargs):
        pass

    @classmethod
    @abc.abstractmethod
    def get_sample_native_cluster_specification(cls, k8_runtime=''):
        pass



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from container_service_extension.common.constants.server_constants import FlattenedClusterSpecKey2X  # noqa: E501
from container_service_extension.common.constants.server_constants import VALID_UPDATE_FIELDS_2X  # noqa: E501
from container_service_extension.exception.exceptions import BadRequestError
from container_service_extension.lib.cloudapi.cloudapi_client import CloudApiClient  # noqa: E501
from container_service_extension.rde.behaviors.behavior_model import BehaviorOperation  # noqa: E501
from container_service_extension.rde.common.entity_service import DefEntityService  # noqa: E501
import container_service_extension.rde.constants as rde_constants
from container_service_extension.rde.models import rde_factory
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
import container_service_extension.rde.models.rde_2_0_0 as rde_2_0_0
import container_service_extension.rde.utils as rde_utils
from container_service_extension.rde.validators.abstract_validator import AbstractValidator  # noqa: E501


class Validator_2_0_0(AbstractValidator):
    def __init__(self):
        pass

    def validate(self, cloudapi_client: CloudApiClient, entity_id: str = None,
                 entity: dict = None, operation: BehaviorOperation = None) -> bool:  # noqa: E501
        """Validate the input request.

        This method performs
        1. Basic validation of the entity by simply casting the input entity
        dict to the model class dictated by the api_version specified in the
        request. This is usually performed for the "create" operation.
        2. Operation (create, update, delete) specific validation.
        - create: "entity" is the only required parameter.
        - update: both "entity" and "entity_id" are required parameters.
        - delete: "entity_id" is the only required parameter.
        - kubeconfig: "entity_id" is the only required parameter.

        :param cloudapi_client: cloud api client
        :param dict entity: dict form of the native entity to be validated
        :param entity_id: entity id to be validated
        :param BehaviorOperation operation: CSE operation key
        :return: is validation successful or failure
        :rtype: bool
        """
        if not entity_id and not entity:
            raise ValueError('Either entity_id or entity is required to validate.')  # noqa: E501
        entity_svc = DefEntityService(cloudapi_client=cloudapi_client)

        api_version: str = cloudapi_client.get_api_version()
        rde_version_introduced_at_api_version: str = rde_utils.get_rde_version_introduced_at_api_version(api_version)  # noqa: E501

        # TODO Reject the request if payload_version does not match with
        #  either rde_in_use (or) rde_version_introduced_at_api_version

        # Cast the entity to the model class based on the user-specified
        # api_version. This can be considered as a basic request validation.
        # Any operation specific validation is handled further down
        NativeEntityClass: AbstractNativeEntity = rde_factory. \
            get_rde_model(rde_version_introduced_at_api_version)
        input_entity = None
        if entity:
            try:
                input_entity: AbstractNativeEntity = NativeEntityClass.from_dict(entity)  # noqa: E501
            except Exception as err:
                msg = f"Failed to parse request body: {err}"
                raise BadRequestError(msg)

        # Return True if the operation is not specified.
        if not operation:
            return True

        # TODO: validators for rest of the CSE operations in V36 will be
        #  implemented as and when v36/def_cluster_handler.py get other handler
        #  functions
        if operation == BehaviorOperation.UPDATE_CLUSTER:
            if not entity_id or not entity:
                raise ValueError('Both entity_id and entity are required to validate the Update operation.')  # noqa: E501
            current_entity: AbstractNativeEntity = entity_svc.get_entity(entity_id).entity  # noqa: E501
            input_entity_spec: rde_2_0_0.ClusterSpec = input_entity.spec
            current_entity_status: rde_2_0_0.Status = current_entity.status
            current_entity_spec = \
                rde_utils.construct_cluster_spec_from_entity_status(
                    current_entity_status, rde_constants.RDEVersion.RDE_2_0_0.value)  # noqa: E501
            return validate_cluster_update_request_and_check_cluster_upgrade(
                input_entity_spec,
                current_entity_spec)

        # TODO check the reason why there was an unreachable raise statement
        raise NotImplementedError(f"Validator for {operation.name} not found")


def validate_cluster_update_request_and_check_cluster_upgrade(input_spec: rde_2_0_0.ClusterSpec,  # noqa: E501
                                                              reference_spec: rde_2_0_0.ClusterSpec) -> bool:  # noqa: E501
    """Validate the desired spec with curr spec and check if upgrade operation.

    :param dict input_spec: input spec
    :param dict reference_spec: reference spec to validate the desired spec
    :return: true if cluster operation is upgrade and false if operation is
        resize
    :rtype: bool
    :raises: BadRequestError for invalid payload.
    """
    exclude_fields = []
    if reference_spec.topology.workers.count == 0:
        # Exclude worker nodes' sizing class and storage profile from
        # validation if worker count is 0
        exclude_fields.append(FlattenedClusterSpecKey2X.WORKERS_SIZING_CLASS.value)  # noqa: E501
        exclude_fields.append(FlattenedClusterSpecKey2X.WORKERS_STORAGE_PROFILE.value)  # noqa: E501
    if reference_spec.topology.nfs.count == 0:
        # Exclude nfs nodes' sizing class and storage profile from validation
        # if nfs count is 0
        exclude_fields.append(FlattenedClusterSpecKey2X.NFS_SIZING_CLASS.value)
        exclude_fields.append(FlattenedClusterSpecKey2X.NFS_STORAGE_PROFILE.value)  # noqa: E501

    input_spec_dict = input_spec.to_dict()
    reference_spec_dict = reference_spec.to_dict()
    diff_fields = rde_utils.find_diff_fields(
        input_spec_dict,
        reference_spec_dict,
        exclude_fields=exclude_fields)

    # Raise exception if empty diff
    if not diff_fields:
        raise BadRequestError("No change in cluster specification")  # noqa: E501

    keys_with_invalid_value = {}
    for k, v in diff_fields.items():
        if k not in VALID_UPDATE_FIELDS_2X:
            keys_with_invalid_value[k] = v

    # Raise exception if fields which cannot be changed are updated
    if len(keys_with_invalid_value) > 0:
        err_msg = "Change detected in immutable field(s) ["
        for k in sorted(keys_with_invalid_value):
            err_msg += \
                f"{k} found : {keys_with_invalid_value[k]['actual']} " \
                f"expected : {keys_with_invalid_value[k]['expected']}, "
        err_msg += "]."
        raise BadRequestError(err_msg)

    is_resize_operation = False
    if FlattenedClusterSpecKey2X.WORKERS_COUNT.value in diff_fields or \
            FlattenedClusterSpecKey2X.NFS_COUNT.value in diff_fields:
        is_resize_operation = True
    is_upgrade_operation = False

    if FlattenedClusterSpecKey2X.TEMPLATE_NAME.value in diff_fields or \
            FlattenedClusterSpecKey2X.TEMPLATE_REVISION.value in diff_fields:
        is_upgrade_operation = True

    # Raise exception if resize and upgrade are performed at the same time
    if is_resize_operation and is_upgrade_operation:
        err_msg = "Cannot resize and upgrade the cluster at the same time"
        raise BadRequestError(err_msg)

    return is_upgrade_operation



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from enum import Enum, unique

NCP_BOUNDARY_FIREWALL_SECTION_NAME = "NCP_BOUNDARY_FIREWALL_SECTION"
NCP_BOUNDARY_TOP_FIREWALL_SECTION_NAME = "NCP_BOUNDARY_TOP_FIREWALL_SECTION"
NCP_BOUNDARY_BOTTOM_FIREWALL_SECTION_NAME = \
    "NCP_BOUNDARY_BOTTOM_FIREWALL_SECTION"

ALL_NODES_IP_SET_NAME = 'ALL_NODES'
ALL_PODS_IP_SET_NAME = 'ALL_PODS'

ALL_NODES_PODS_NSGROUP_NAME = "ALL_NODES_PODS"

# Prefix for Gateway urn
GATEWAY_URN_PREFIX = 'urn:vcloud:gateway'

# NSXT Gateway NAT constants
NATS_PATH_FRAGMENT = 'nat'
RULES_PATH_FRAGMENT = 'rules'
DNAT_RULE_TYPE = 'DNAT'

# NSX-T constants
NSXT_BACKED_GATEWAY_UPLINK_INDEX = 0  # NSX-T gateway only have 1 ext network
NSXT_PUT_REQUEST_WAIT_TIME = 3

# External network available ip address path fragment
AVAILABLE_IP_PATH_FRAGMENT = 'availableIpAddresses'

# NSX-T realized status
NSXT_GATEWAY_REALIZED_STATUS = 'REALIZED'


class RequestMethodVerb(Enum):
    GET = 'Get'
    POST = 'Post'
    PUT = 'Put'
    DELETE = 'Delete'


class INSERT_POLICY(Enum):
    INSERT_TOP = 'insert_top'
    INSERT_BEFORE = 'insert_before'
    INSERT_AFTER = 'insert_after'
    INSERT_BOTTOM = 'insert_bottom'


class FIREWALL_ACTION(Enum):
    ALLOW = "ALLOW"
    DROP = "DROP"
    REJECT = "REJECT"


@unique
class NsxtNATRuleKey(str, Enum):
    """Keys for NSXT NAT Rules."""

    NAME = 'name'
    DESCRIPTION = 'description'
    ENABLED = 'enabled'
    RULE_TYPE = 'ruleType'
    EXTERNAL_ADDRESSES = 'externalAddresses'
    INTERNAL_ADDRESSES = 'internalAddresses'
    LOGGING = 'logging'
    APPLICATION_PORT_PROFILE = 'applicationPortProfile'
    DNAT_EXTERNAL_PORT = 'dnatExternalPort'
    ID = 'id'


@unique
class NsxtGatewayRequestKey(str, Enum):
    EDGE_GATEWAY_UPLINKS = 'edgeGatewayUplinks'
    UPLINK_ID = 'uplinkId'
    GATEWAY = 'gateway'
    PREFIX_LENGTH = 'prefixLength'
    TOTAL_IP_COUNT = 'totalIpCount'
    AUTO_ALLOCATE_IP_RANGES = 'autoAllocateIpRanges'
    SUBNETS = 'subnets'
    VALUES = 'values'
    IP_RANGES = 'ipRanges'
    START_ADDRESS = 'startAddress'
    END_ADDRESS = 'endAddress'
    STATUS = 'status'
    IP_ADDRESS = 'ipAddress'



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from typing import List, Optional

import lxml.objectify as lxml
import pyvcloud.vcd.client as vcd_client
import pyvcloud.vcd.org as vcd_org
import pyvcloud.vcd.role as vcd_role

import container_service_extension.lib.cloudapi.cloudapi_client as cloudapi_client  # noqa: E501
import container_service_extension.logging.logger as logger

ORG_ADMIN_RIGHTS = [
    'General: Administrator Control',
    'General: Administrator View'
]


class UserContext:
    def __init__(self, client: vcd_client.Client,
                 cloud_api_client: cloudapi_client.CloudApiClient):
        self.client: vcd_client.Client = client
        self.cloud_api_client: cloudapi_client.CloudApiClient = \
            cloud_api_client
        self._session: Optional[lxml.ObjectifiedElement] = None
        self._name: Optional[str] = None
        self._id: Optional[str] = None
        self._org_name: Optional[str] = None
        self._org_href: Optional[str] = None
        self._role: Optional[str] = None
        self._rights: Optional[List[str]] = None

    @property
    def session(self):
        if self._session is None:
            self._session = self.client.get_vcloud_session()
        return self._session

    @property
    def name(self):
        """User name from vCD session."""
        if self._name is None:
            self._name = self.session.get('user')
        return self._name

    @property
    def id(self):
        """User ID from vCD session."""
        if self._id is None:
            self._id = self.session.get('userId')
        return self._id

    @property
    def org_name(self):
        if self._org_name is None:
            self._org_name = self.session.get('org')
        return self._org_name

    @property
    def org_href(self):
        if self._org_href is None:
            self._org_href = self.client.get_org().get('href')
        return self._org_href

    @property
    def role(self):
        if self._role is None:
            self._role = self.session.get('roles')
        return self._role

    @property
    def rights(self):
        # If the user is unable to fetch the list of rights, then they
        # can't be sys admin or org admin. And anyone other than those
        # two shouldn't be able to see their own set of rights.
        if self._rights is None:
            try:
                org = vcd_org.Org(self.client, href=self.org_href)
                role = vcd_role.Role(self.client,
                                     resource=org.get_role_resource(self.role))

                self._rights = []
                for right_dict in role.list_rights():
                    right_name = right_dict.get('name')
                    if right_name is not None:
                        self._rights.append(right_name)
            # maybe replace with Forbidden exception?
            except Exception:
                msg = f"Unable to fetch right records for User: {self.name}"
                logger.SERVER_LOGGER.debug(msg, exc_info=True)

        return self._rights

    @property
    def has_org_admin_rights(self):
        return all(right in self.rights for right in ORG_ADMIN_RIGHTS)



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
from dataclasses import asdict
import functools
import json
from typing import Iterator, List

from requests.exceptions import HTTPError

import container_service_extension.common.constants.shared_constants as cse_shared_constants  # noqa: E501
import container_service_extension.exception.exceptions as cse_exceptions
from container_service_extension.exception.minor_error_codes import MinorErrorCode  # noqa: E501
from container_service_extension.lib.cloudapi.cloudapi_client import CloudApiClient  # noqa: E501
from container_service_extension.lib.cloudapi.constants import CloudApiResource
from container_service_extension.lib.cloudapi.constants import CloudApiVersion
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
from container_service_extension.rde.behaviors.behavior_model import Behavior, BehaviorAclEntry  # noqa: E501
import container_service_extension.rde.utils as def_utils


def handle_behavior_service_exception(func):
    """Decorate to trap exceptions and process them.

    Raise errors of type HTTPError as DefSchemaServiceError.
    Re-raise any other exception as it is.

    This decorator should be applied only on method of schema_service.py

    :param method func: decorated function

    :return: reference to the function that executes the decorated function
        and traps exceptions raised by it.
    """
    @functools.wraps(func)
    def exception_handler_wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except HTTPError as error:
            response_dict = json.loads(error.response.text)
            error_message = response_dict.get('message')
            LOGGER.error(error_message)
            raise cse_exceptions.BehaviorServiceError(
                error_message=error_message,
                minor_error_code=MinorErrorCode.DEFAULT_ERROR_CODE)
        except Exception as error:
            LOGGER.error(error)
            raise error
        return result
    return exception_handler_wrapper


class BehaviorService:
    def __init__(self, cloudapi_client: CloudApiClient):
        def_utils.raise_error_if_def_not_supported(cloudapi_client)
        self._cloudapi_client = cloudapi_client

    @handle_behavior_service_exception
    def create_behavior_on_interface(self, behavior: Behavior, interface_id) -> Behavior:  # noqa: E501
        """Create Behavior on the specified Interface.

        :param behavior: Behavior to be created.
        :param interface_id: Interface Id on which behavior is supposed to be
        created.
        :return: The created behavior object.
        """
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.POST,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.INTERFACES}"
                                       f"/{interface_id}"
                                       f"/{CloudApiResource.BEHAVIORS}",
            payload=asdict(behavior))
        return Behavior(**response_body)

    @handle_behavior_service_exception
    def update_behavior_on_interface(self, behavior: Behavior, interface_id) -> Behavior:  # noqa: E501
        """Update the behavior on the specified interface Id.

        :param behavior: Behavior to be updated.
        :param interface_id: Interface Id on which the behavior to be updated
        is present.
        :return: The updated behavior object.
        """
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.PUT,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.INTERFACES}"
                                       f"/{interface_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior.id}",
            payload=asdict(behavior))
        return Behavior(**response_body)

    @handle_behavior_service_exception
    def list_behaviors_on_interface(self, interface_id) -> Iterator[Behavior]:
        """List all the behaviors on the specified interface.

        :param interface_id: Interface Id.
        :return: List of behaviors on the interface.
        """
        page_num = 0
        while True:
            page_num += 1
            response_body = self._cloudapi_client.do_request(
                method=cse_shared_constants.RequestMethod.GET,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=f"{CloudApiResource.INTERFACES}"
                                           f"/{interface_id}"
                                           f"/{CloudApiResource.BEHAVIORS}?"
                                           f"page={page_num}")
            if len(response_body['values']) > 0:
                for behavior in response_body['values']:
                    yield Behavior(**behavior)
            else:
                break

    @handle_behavior_service_exception
    def get_behavior_on_interface_by_id(self, behavior_id, interface_id) -> Behavior:  # noqa: E501
        """Get the behavior details by its ID on the given interface.

        :param behavior_id: Id of the behavior.
        :param interface_id: Id of the interface.
        :return: The behavior details.
        """
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.INTERFACES}"
                                       f"/{interface_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior_id}")
        return Behavior(**response_body)

    @handle_behavior_service_exception
    def get_behavior_on_interface_by_name(self, behavior_name, interface_id) -> Behavior:  # noqa: E501
        """Get the behavior details by its name on the given interface.

        :param behavior_name: Name of the behavior.
        :param interface_id: Id of the interface.
        :return: The behavior details.
        """
        # TODO Check with Extensibility team to ensure there canno exist
        #  duplicates.
        behaviors = self.list_behaviors_on_interface(interface_id=interface_id)
        for behavior in behaviors:
            if behavior.name == behavior_name:
                return behavior

    @handle_behavior_service_exception
    def delete_behavior_on_interface(self, behavior_id: str, interface_id: str):  # noqa: E501
        """Delete the behavior by its Id on the specified interface.

        :param behavior_id: Id of the behavior.
        :param interface_id: Id of the interface.
        """
        if not self._cloudapi_client.is_sys_admin:
            raise ValueError("Cloud API Client should be sysadmin.")
        return self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.DELETE,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.INTERFACES}"
                                       f"/{interface_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior_id}")

    @handle_behavior_service_exception
    def create_behavior_on_entity_type(self, behavior: Behavior, entity_type_id) -> Behavior:  # noqa: E501
        """Create a behavior directly on the entity type.

        This method is non-functional. Under discussion with the Extensibility
        team on the error.
        TODO: Either update the docstring (or) remove the method.

        :param behavior:
        :param entity_type_id:
        :return:
        """
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.POST,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                       f"/{entity_type_id}"
                                       f"/{CloudApiResource.BEHAVIORS}",
            payload=asdict(behavior))
        return Behavior(**response_body)

    @handle_behavior_service_exception
    def override_behavior_on_entity_type(self, behavior: Behavior, entity_type_id) -> Behavior:  # noqa: E501
        """Override the behavior-interface on the specified entity type.

        Only Execution portion of the Behavior can be overridden.

        :param behavior: The updated behavior to be overridden on the entity-type.  # noqa: E501
        :param entity_type_id: Id of the entity type
        :return: Overridden behavior on the entity type.
        """
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.PUT,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                       f"/{entity_type_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior.ref}",
            payload=asdict(behavior))
        return Behavior(**response_body)

    @handle_behavior_service_exception
    def list_behaviors_on_entity_type(self, entity_type_id) -> Iterator[Behavior]:  # noqa: E501
        """List behaviors on the specified entity type.

        :param entity_type_id: Id of the entity type.
        :return: List of behaviors.
        """
        # TODO Test this later, there is pagination issue on the entity types
        #  endpoint.
        page_num = 0
        while True:
            page_num += 1
            response_body = self._cloudapi_client.do_request(
                method=cse_shared_constants.RequestMethod.GET,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                           f"/{entity_type_id}"
                                           f"/{CloudApiResource.BEHAVIORS}?"
                                           f"page={page_num}")
            if len(response_body['values']) > 0:
                for behavior in response_body['values']:
                    yield Behavior(**behavior)
            else:
                break

    @handle_behavior_service_exception
    def get_behavior_on_entity_type_by_id(self, behavior_interface_id, entity_type_id) -> Behavior:  # noqa: E501
        """Get the behavior on the entity type id.

        Note: Passing behavior-type id will not be accepted.

        :param behavior_interface_id: Id of the behavior-interface.
        :param entity_type_id: Id of the entity type.
        :return: The behavior details.
        """
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                       f"/{entity_type_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior_interface_id}")
        return Behavior(**response_body)

    @handle_behavior_service_exception
    def get_behavior_on_entity_type_by_name(self, behavior_name, entity_type_id) -> Behavior:  # noqa: E501
        """Get the behavior by it's name on the specified entity type.

        :param behavior_name: Name of the behavior
        :param entity_type_id: Id of the entity type.
        :return: Behavior details
        """
        # TODO Test this later, there is pagination issue on the entity types
        #  endpoint. Ensure there cannot exist duplicates.
        behaviors = self.list_behaviors_on_entity_type(entity_type_id=entity_type_id)  # noqa: E501
        for behavior in behaviors:
            if behavior.name == behavior_name:
                return behavior

    @handle_behavior_service_exception
    def delete_behavior_on_entity_type(self, behavior_interface_id, entity_type_id):  # noqa: E501
        """Delete the behavior on the entity type.

        Note: Passing behavior-type id will not be accepted.

        :param behavior_interface_id: Id of the behavior interface.
        :param entity_type_id: Id of the entity type.
        """
        if not self._cloudapi_client.is_sys_admin:
            raise ValueError("Cloud API Client should be sysadmin.")
        return self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.DELETE,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                       f"/{entity_type_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior_interface_id}")

    @handle_behavior_service_exception
    def list_behavior_acls_on_entity_type(self, entity_type_id: str) -> Iterator[BehaviorAclEntry]:  # noqa: E501
        """Get the list of access controls of the individual behaviors.

        Get the list of access controls of the individual behaviors on the
        specified entity type.

        :param entity_type_id: Id of the entity type.
        :return: List of Behavior access controls.
        """
        page_num = 0
        while True:
            page_num += 1
            response_body = self._cloudapi_client.do_request(
                method=cse_shared_constants.RequestMethod.GET,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                           f"/{entity_type_id}"
                                           f"/{CloudApiResource.BEHAVIOR_ACLS}")  # noqa: E501
            if len(response_body['values']) > 0:
                for acl in response_body['values']:
                    yield BehaviorAclEntry(**acl)
            else:
                break

    @handle_behavior_service_exception
    def update_behavior_acls_on_entity_type(self, entity_type_id: str,
                                            behavior_acl_list: List[BehaviorAclEntry]):  # noqa: E501
        """Update the list of behavior access controls.

        Update the list of behavior access controls on the specified entity
        type. It is strictly recommended to retrieve the current list before
        updating it. Any missing records in the payload will be deleted.

        :param entity_type_id: Id of the entity type.
        :param behavior_acl_list: The updated list of behavior access controls.
        :return: List of updated behavior access controls.
        """
        if not self._cloudapi_client.is_sys_admin:
            raise ValueError("Cloud API Client should be sysadmin.")
        response_body = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.PUT,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITY_TYPES}"
                                       f"/{entity_type_id}"
                                       f"/{CloudApiResource.BEHAVIOR_ACLS}",
            payload={"values": [asdict(acl) for acl in behavior_acl_list]})
        return [BehaviorAclEntry(**acl) for acl in response_body['values']]

    @handle_behavior_service_exception
    def invoke_behavior(self, entity_id: str, behavior_interface_id: str,
                        arguments: dict = None) -> str:
        """Invoke behavior on the RDE.

        Below is the optional sample payload for the behavior invocation.
        {
            "arguments": {
                "argument1": "argument1"
            }
        }
        :param entity_id: Id of the entity
        :param behavior_interface_id: Id of the behavior interface
        :param arguments: Arguments to the behavior
        :return: task href in the Location header
        :rtype: str
        """
        # TODO Invocation of the behavior must return the taskID retrieved from
        #  the response headers. This is yet to be done by Extensibility team.
        payload = None
        if arguments:
            payload = {
                "arguments": arguments
            }
        _, response_headers = self._cloudapi_client.do_request(
            method=cse_shared_constants.RequestMethod.POST,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}"
                                       f"/{entity_id}"
                                       f"/{CloudApiResource.BEHAVIORS}"
                                       f"/{behavior_interface_id}"
                                       f"/{CloudApiResource.BEHAVIOR_INVOCATION}",  # noqa: E501
            payload=payload,
            return_response_headers=True)
        return response_headers[cse_shared_constants.HttpResponseHeader.LOCATION]  # noqa: E501



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
import copy

import pyvcloud.vcd.client as vcd_client

from container_service_extension.common.constants.server_constants import K8S_PROVIDER_KEY  # noqa: E501
from container_service_extension.common.constants.server_constants import K8sProvider  # noqa: E501
from container_service_extension.common.constants.server_constants import PKS_CLUSTER_DOMAIN_KEY  # noqa: E501
from container_service_extension.common.constants.server_constants import PKS_PLANS_KEY  # noqa: E501
from container_service_extension.common.constants.server_constants import ThreadLocalData  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
import container_service_extension.common.thread_local_data as thread_local_data  # noqa: E501
import container_service_extension.common.utils.ovdc_utils as ovdc_utils
import container_service_extension.common.utils.server_utils as server_utils
from container_service_extension.exception.exceptions import ClusterAlreadyExistsError  # noqa: E501
from container_service_extension.exception.exceptions import ClusterNotFoundError  # noqa: E501
from container_service_extension.exception.exceptions import CseServerError
from container_service_extension.exception.exceptions import PksClusterNotFoundError  # noqa: E501
from container_service_extension.exception.exceptions import PksDuplicateClusterError  # noqa: E501
from container_service_extension.exception.exceptions import PksServerError
from container_service_extension.lib.telemetry.constants import CseOperation
from container_service_extension.lib.telemetry.constants import PayloadKey
import container_service_extension.lib.telemetry.telemetry_handler as telemetry_handler  # noqa: E501
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
import container_service_extension.security.context.operation_context as ctx
from container_service_extension.server.pks.pksbroker import PksBroker
import container_service_extension.server.pks.pksbroker_manager as pks_broker_manager  # noqa: E501
from container_service_extension.server.pks.pksbroker_manager import create_pks_context_for_all_accounts_in_org  # noqa: E501
import container_service_extension.server.request_handlers.request_utils as req_utils  # noqa: E501

DEFAULT_API_VERSION = vcd_client.ApiVersion.VERSION_33.value


@telemetry_handler.record_user_action_telemetry(cse_operation=CseOperation.PKS_CLUSTER_LIST)  # noqa: E501
def cluster_list(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    All brokers in the org do 'list cluster' operation.
    Post-process the result returned by the broker.
    Aggregate all the results into a list.

    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: List
    """
    _raise_error_if_pks_not_enabled()

    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])

    cse_params = copy.deepcopy(data)
    cse_params[PayloadKey.SOURCE_DESCRIPTION] = thread_local_data.get_thread_local_data(ThreadLocalData.USER_AGENT)  # noqa: E501
    telemetry_handler.record_user_action_details(
        cse_operation=CseOperation.PKS_CLUSTER_LIST,
        cse_params=cse_params)

    pks_clusters_info = pks_broker_manager.list_clusters(data, op_ctx)
    common_cluster_properties = [
        'name',
        'vdc',
        'status',
        'org_name',
        'k8s_version',
        K8S_PROVIDER_KEY
    ]

    result = []
    for info in pks_clusters_info:
        filtered_cluster_info = \
            {k: info.get(k) for k in common_cluster_properties}
        result.append(filtered_cluster_info)

    return result


@telemetry_handler.record_user_action_telemetry(cse_operation=CseOperation.PKS_CLUSTER_INFO)  # noqa: E501
def cluster_info(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster info operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    _raise_error_if_pks_not_enabled()

    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])

    cluster, _ = _get_cluster_info(data, op_ctx)  # noqa: E501
    telemetry_handler.record_user_action_details(
        cse_operation=CseOperation.PKS_CLUSTER_INFO,
        cse_params=_get_telemetry_data(data, cluster))
    return cluster


@telemetry_handler.record_user_action_telemetry(cse_operation=CseOperation.PKS_CLUSTER_CONFIG)  # noqa: E501
def cluster_config(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster config operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    _raise_error_if_pks_not_enabled()
    cluster, broker = _get_cluster_info(request_data, op_ctx, telemetry=False)  # noqa: E501
    telemetry_handler.record_user_action_details(
        cse_operation=CseOperation.PKS_CLUSTER_CONFIG,
        cse_params=_get_telemetry_data(request_data, cluster))
    return broker.get_cluster_config(data=request_data)


@telemetry_handler.record_user_action_telemetry(cse_operation=CseOperation.PKS_CLUSTER_CREATE)  # noqa: E501
def cluster_create(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster create operation.

    Required data: org_name, ovdc_name, cluster_name

    (data validation handled in broker)

    :return: Dict
    """
    _raise_error_if_pks_not_enabled()

    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])

    required = [
        RequestKey.CLUSTER_NAME
    ]
    req_utils.validate_payload(data, required)

    cluster_name = data[RequestKey.CLUSTER_NAME]
    data['is_org_admin_search'] = True

    try:
        _get_cluster_and_broker(data, op_ctx, telemetry=False)
        raise ClusterAlreadyExistsError(f"Cluster {cluster_name} "
                                        f"already exists.")
    except ClusterNotFoundError:
        pass

    sysadmin_client_v33 = op_ctx.get_sysadmin_client(
        api_version=DEFAULT_API_VERSION)
    k8s_metadata = \
        ovdc_utils.get_ovdc_k8s_provider_metadata(
            sysadmin_client_v33,
            org_name=data[RequestKey.ORG_NAME],
            ovdc_name=data[RequestKey.OVDC_NAME],
            include_credentials=True,
            include_nsxt_info=True)
    broker = _get_broker_from_k8s_metadata(k8s_metadata, op_ctx)
    data[RequestKey.PKS_PLAN_NAME] = k8s_metadata[PKS_PLANS_KEY][0]
    data[RequestKey.PKS_EXT_HOST] = \
        f"{cluster_name}.{k8s_metadata[PKS_CLUSTER_DOMAIN_KEY]}"
    cluster = broker.create_cluster(data=data)
    # Record telemetry data
    telemetry_handler.record_user_action_details(
        cse_operation=CseOperation.PKS_CLUSTER_CREATE,
        cse_params=_get_telemetry_data(data, cluster))
    return cluster


@telemetry_handler.record_user_action_telemetry(cse_operation=CseOperation.PKS_CLUSTER_DELETE)  # noqa: E501
def cluster_delete(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster delete operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    _raise_error_if_pks_not_enabled()
    cluster, broker = _get_cluster_info(request_data, op_ctx, telemetry=False)  # noqa: E501
    # Record telemetry data
    telemetry_handler.record_user_action_details(
        cse_operation=CseOperation.PKS_CLUSTER_DELETE,
        cse_params=_get_telemetry_data(request_data, cluster))
    return broker.delete_cluster(data=request_data)


@telemetry_handler.record_user_action_telemetry(cse_operation=CseOperation.PKS_CLUSTER_RESIZE)  # noqa: E501
def cluster_resize(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster resize operation.

    Required data: cluster_name, num_nodes
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    _raise_error_if_pks_not_enabled()

    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])

    cluster, broker = _get_cluster_info(data, op_ctx, telemetry=False)  # noqa: E501
    # Record telemetry data
    telemetry_handler.record_user_action_details(
        cse_operation=CseOperation.PKS_CLUSTER_RESIZE,
        cse_params=_get_telemetry_data(data, cluster))
    return broker.resize_cluster(data=data)


def _get_cluster_info(request_data, op_ctx, **kwargs):
    """Get cluster details directly from cloud provider.

    Logic of the method is as follows.

    If 'ovdc' is present in the cluster spec,
        choose the right PKS broker to do get_cluster operation.
    else
        Invoke set of all PKS brokers in the org to find the cluster

    :return: a tuple of cluster information as dictionary and the broker
        instance used to find the cluster information.

    :rtype: tuple
    """
    required = [
        RequestKey.CLUSTER_NAME
    ]
    req_utils.validate_payload(request_data, required)

    org_name = request_data.get(RequestKey.ORG_NAME)
    ovdc_name = request_data.get(RequestKey.OVDC_NAME)

    if ovdc_name is not None and org_name is not None:
        sysadmin_client_v33 = op_ctx.get_sysadmin_client(
            api_version=DEFAULT_API_VERSION)
        k8s_metadata = \
            ovdc_utils.get_ovdc_k8s_provider_metadata(
                sysadmin_client_v33,
                org_name=org_name,
                ovdc_name=ovdc_name,
                include_credentials=True,
                include_nsxt_info=True)
        broker = _get_broker_from_k8s_metadata(
            k8s_metadata, op_ctx)
        return broker.get_cluster_info(data=request_data, **kwargs), broker

    return _get_cluster_and_broker(request_data, op_ctx, **kwargs)


def _get_cluster_and_broker(request_data, op_ctx, **kwargs):
    cluster_name = request_data[RequestKey.CLUSTER_NAME]

    pks_ctx_list = create_pks_context_for_all_accounts_in_org(op_ctx)
    for pks_ctx in pks_ctx_list:
        debug_msg = f"Get cluster info for cluster '{cluster_name}' " \
            f"failed on host '{pks_ctx['host']}' with error: "
        pks_broker = PksBroker(pks_ctx, op_ctx)
        try:
            return pks_broker.get_cluster_info(data=request_data, **kwargs), pks_broker  # noqa: E501
        except (PksClusterNotFoundError, PksServerError) as err:
            # continue searching using other PksBrokers
            LOGGER.debug(f"{debug_msg}{err}")
        except PksDuplicateClusterError as err:
            # fail because multiple clusters with same name exist
            LOGGER.debug(f"{debug_msg}{err}")
            raise
        except Exception as err:
            LOGGER.error(f"Unknown error: {err}", exc_info=True)
            raise

    # raised if cluster was not found in PksBrokers
    raise ClusterNotFoundError(f"Cluster '{cluster_name}' not found.")


def _get_broker_from_k8s_metadata(k8s_metadata,
                                  op_ctx: ctx.OperationContext):
    """Get broker from ovdc k8s metadata.

    If PKS is not enabled, raise CseServerError
    If PKS is enabled
        if no ovdc metadata exists or k8s provider is None, raise server error
        else return the broker according to ovdc k8s provider
    """
    _raise_error_if_pks_not_enabled()
    if not k8s_metadata or k8s_metadata.get(K8S_PROVIDER_KEY) != K8sProvider.PKS:  # noqa: E501
        raise CseServerError("Org VDC is not enabled for Kubernetes "
                             "cluster deployment")
    return PksBroker(k8s_metadata, op_ctx)


def _raise_error_if_pks_not_enabled():
    if not server_utils.is_pks_enabled():
        raise CseServerError('CSE is not configured to work with PKS.')


def _get_telemetry_data(request_data, cluster_data):
    """Construct telemetry data.

    :param dict request_data: data from user
    :param dict cluster_data: cluster information
    :return: telemetry data
    :rtype: dict
    """
    cse_params = copy.deepcopy(request_data)
    cse_params[PayloadKey.K8S_VERSION] = cluster_data.get(PayloadKey.K8S_VERSION)  # noqa: E501
    cse_params[PayloadKey.PKS_CLUSTER_ID] = cluster_data.get(PayloadKey.PKS_CLUSTER_ID)  # noqa: E501
    cse_params[PayloadKey.PKS_VERSION] = cluster_data.get(PayloadKey.PKS_VERSION)  # noqa: E501
    cse_params[PayloadKey.SOURCE_DESCRIPTION] = thread_local_data.get_thread_local_data(ThreadLocalData.USER_AGENT)  # noqa: E501
    return cse_params



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Utility methods to help interaction with defined entities framework."""
import importlib
from importlib import resources as pkg_resources
import json
import math
from typing import Type
from typing import Union

import semantic_version

import container_service_extension.common.utils.core_utils as core_utils
import container_service_extension.common.utils.server_utils as server_utils
import container_service_extension.exception.exceptions as exceptions
from container_service_extension.lib.cloudapi.cloudapi_client import CloudApiClient  # noqa: E501
import container_service_extension.rde.constants as def_constants
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
import container_service_extension.rde.models.rde_1_0_0 as rde_1_0_0
import container_service_extension.rde.models.rde_2_0_0 as rde_2_0_0
import container_service_extension.rde.models.rde_factory as rde_factory


def raise_error_if_def_not_supported(cloudapi_client: CloudApiClient):
    """Raise DefNotSupportedException if defined entities are not supported.

    :param cloudapi_client CloudApiClient
    """
    if float(cloudapi_client.get_api_version()) < \
            def_constants.DEF_API_MIN_VERSION:
        raise exceptions.DefNotSupportedException(
            "Defined entity framework is not"
            " supported for {cloudapi_client.get_api_version()}")


def generate_interface_id(vendor, nss, version):
    """Generate defined entity interface id.

    By no means, id generation in this method, guarantees the actual
    entity type registration with vCD.

    :param vendor (str)
    :param nss (str)
    :param version (str)

    :rtype str
    """
    return f"{def_constants.DEF_INTERFACE_ID_PREFIX}:{vendor}:{nss}:{version}"


def generate_entity_type_id(vendor, nss, version):
    """Generate defined entity type id.

    By no means, id generation in this method, guarantees the actual
    interface registration with vCD.

    :param vendor (str)
    :param nss (str)
    :param version (str)

    :rtype str
    """
    return f"{def_constants.DEF_ENTITY_TYPE_ID_PREFIX}:{vendor}:{nss}:{version}"  # noqa: E501


def get_runtime_rde_version_by_vcd_api_version(vcd_api_version: str) -> str:
    major_vcd_api_version = str(math.floor(float(vcd_api_version)) * 1.0)
    val = def_constants.MAP_VCD_API_VERSION_TO_RUNTIME_RDE_VERSION.get(major_vcd_api_version)  # noqa: E501
    if not val:
        val = '0.0.0'
    return val


def get_rde_version_introduced_at_api_version(vcd_api_version: str) -> str:
    return def_constants.MAP_VCD_API_VERSION_TO_RDE_VERSION[vcd_api_version]


def get_rde_metadata(rde_version: str) -> dict:
    from container_service_extension.rde.models.common_models import MAP_RDE_VERSION_TO_ITS_METADATA  # noqa: E501
    return MAP_RDE_VERSION_TO_ITS_METADATA[rde_version]


def construct_cluster_spec_from_entity_status(entity_status: Union[rde_1_0_0.Status, rde_2_0_0.Status], rde_version: str) -> Union[rde_1_0_0.ClusterSpec, rde_2_0_0.ClusterSpec]:  # noqa: E501
    """Construct cluster spec of the specified rde version from the current status of the cluster.

    :param rde_X_X_X Status entity_status: Entity Status of rde of given version  # noqa: E501
    :param str rde_version: RDE version of the spec to be created from the status # noqa: E501
    :return: Cluster Specification of respective rde_version_in_use
    :raises NotImplementedError
    """
    # TODO: Refactor this multiple if to rde_version -> handler pattern
    if rde_version == def_constants.RDEVersion.RDE_2_0_0.value:
        return construct_2_0_0_cluster_spec_from_entity_status(entity_status)
    # elif rde_version == def_constants.RDEVersion.RDE_2_1_0:
    #    return construct_2_1_0_cluster_spec_from_entity_status(entity_status)
    raise NotImplementedError(f"constructing cluster spec from entity status for {rde_version} is"  # noqa:
                              f" not implemented ")


def construct_2_0_0_cluster_spec_from_entity_status(entity_status: rde_2_0_0.Status) -> rde_2_0_0.ClusterSpec:  # noqa:
    """Construct cluster specification from entity status using rde_2_0_0 model.

    :param rde_2_0_0.Status entity_status: Entity Status as defined in rde_2_0_0  # noqa: E501
    :return: Cluster Specification as defined in rde_2_0_0 model
    """
    # Currently only single control-plane is supported.
    control_plane = rde_2_0_0.ControlPlane(
        sizing_class=entity_status.nodes.control_plane.sizing_class,
        storage_profile=entity_status.nodes.control_plane.storage_profile,
        count=1)

    workers_count = len(entity_status.nodes.workers)
    if workers_count == 0:
        workers = rde_2_0_0.Workers(sizing_class=None,
                                    storage_profile=None,
                                    count=0)
    else:
        workers = rde_2_0_0.Workers(
            sizing_class=entity_status.nodes.workers[0].sizing_class,
            storage_profile=entity_status.nodes.workers[0].storage_profile,
            count=workers_count)

    nfs_count = len(entity_status.nodes.nfs)
    if nfs_count == 0:
        nfs = rde_2_0_0.Nfs(sizing_class=None, storage_profile=None,
                            count=0)
    else:
        nfs = rde_2_0_0.Nfs(
            sizing_class=entity_status.nodes.nfs[0].sizing_class,  # noqa: E501
            storage_profile=entity_status.nodes.nfs[
                0].storage_profile, count=nfs_count)

    k8_distribution = rde_2_0_0.Distribution(
        template_name=entity_status.cloud_properties.distribution.template_name,  # noqa: E501
        template_revision=entity_status.cloud_properties.distribution.template_revision)  # noqa: E501

    network_settings = rde_2_0_0.Network(
        expose=entity_status.cloud_properties.exposed)

    settings = rde_2_0_0.Settings(
        ovdc_network=entity_status.cloud_properties.ovdc_network_name,
        ssh_key=entity_status.cloud_properties.ssh_key,
        rollback_on_failure=entity_status.cloud_properties.rollback_on_failure,  # noqa: E501
        network=network_settings)

    topology = rde_2_0_0.Topology(
        control_plane=control_plane,
        workers=workers,
        nfs=nfs)

    return rde_2_0_0.ClusterSpec(settings=settings,
                                 distribution=k8_distribution,
                                 topology=topology)


def load_rde_schema(schema_file: str) -> dict:
    try:
        schema_module = importlib.import_module(def_constants.DEF_SCHEMA_DIRECTORY)  # noqa: E501
        schema_file = pkg_resources.open_text(schema_module, schema_file)
        return json.load(schema_file)
    except (ImportError, ModuleNotFoundError, FileNotFoundError) as e:
        raise e
    finally:
        try:
            schema_file.close()
        except Exception:
            pass


def find_diff_fields(input_spec: dict, reference_spec: dict, exclude_fields: list = None) -> dict:  # noqa: E501
    if exclude_fields is None:
        exclude_fields = []
    input_dict = core_utils.flatten_dictionary(input_spec)
    reference_dict = core_utils.flatten_dictionary(reference_spec)
    exclude_key_set = set(exclude_fields)
    key_set_for_validation = set(input_dict.keys()) - exclude_key_set
    result = {}
    for key in key_set_for_validation:
        expected = reference_dict.get(key)
        actual = input_dict.get(key)
        if actual is not None and actual != expected:
            result[key] = {
                "expected": expected,
                "actual": actual
            }
    return result


def raise_error_if_unsupported_payload_version(payload_version: str):
    input_rde_version = def_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION.get(payload_version, def_constants.PayloadKey.UNKNOWN)  # noqa: E501
    runtime_rde_version = server_utils.get_rde_version_in_use()
    if input_rde_version == def_constants.PayloadKey.UNKNOWN or semantic_version.Version(input_rde_version) > semantic_version.Version(runtime_rde_version):  # noqa: E501
        raise exceptions.BadRequestError(f"Unsupported payload version: {payload_version}")  # noqa: E501


def convert_input_rde_to_runtime_rde_format(input_entity: dict):
    """Convert input entity to runtime rde format.

    :param AbstractNativeEntity input_entity: entity to be converted
    :return: Entity in runtime rde format
    :rtype: AbstractNativeEntity
    """
    payload_version = input_entity.get(def_constants.PayloadKey.PAYLOAD_VERSION)  # noqa: E501
    if payload_version:
        input_rde_version = def_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION[payload_version]  # noqa: E501
    else:
        input_rde_version = def_constants.RDEVersion.RDE_1_0_0.value
    runtime_rde_version = server_utils.get_rde_version_in_use()
    RuntimeNativeEntityClass: Type[AbstractNativeEntity] = rde_factory.get_rde_model(runtime_rde_version)  # noqa: E501
    InputNativeEntityClass: Type[AbstractNativeEntity] = rde_factory.get_rde_model(input_rde_version)  # noqa: E501
    input_rde: AbstractNativeEntity = InputNativeEntityClass.from_dict(input_entity)  # noqa: E501
    return RuntimeNativeEntityClass.from_native_entity(input_rde)


def convert_runtime_rde_to_input_rde_version_format(runtime_rde: AbstractNativeEntity, input_rde_version):  # noqa: E501
    InputNativeEntityClass: Type[AbstractNativeEntity] = rde_factory.get_rde_model(input_rde_version)  # noqa: E501
    output_entity = InputNativeEntityClass.from_native_entity(runtime_rde)
    return output_entity



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from container_service_extension.common.constants.server_constants import K8S_PROVIDER_KEY  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_DEFAULT_PAGE_SIZE  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_FIRST_PAGE_NUMBER  # noqa: E501
from container_service_extension.common.constants.shared_constants import PaginationKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
import container_service_extension.common.utils.server_utils as server_utils
from container_service_extension.lib.telemetry.constants import CseOperation
from container_service_extension.lib.telemetry.telemetry_handler import \
    record_user_action_telemetry
import container_service_extension.security.context.operation_context as ctx
import container_service_extension.server.request_handlers.request_utils as req_utils  # noqa: E501
from container_service_extension.server.vcdbroker import VcdBroker


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_CREATE)
def cluster_create(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster create operation.

    Required data: org_name, ovdc_name, cluster_name
    Conditional data and default values:
            network_name, num_nodes=2, num_cpu=None, mb_memory=None,
            storage_profile_name=None, template_name=default,
            template_revision=default, ssh_key=None, enable_nfs=False,
            rollback=True

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])
    return vcd_broker.create_cluster(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_RESIZE)
def cluster_resize(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster resize operation.

    Required data: cluster_name, num_nodes
    Optional data and default values: org_name=None, ovdc_name=None
    Conditional data and default values:
            network_name, rollback=True

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])
    return vcd_broker.resize_cluster(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_DELETE)
def cluster_delete(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster delete operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])
    return vcd_broker.delete_cluster(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_INFO)
def cluster_info(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster info operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])
    return vcd_broker.get_cluster_info(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_CONFIG)
def cluster_config(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster config operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])
    return vcd_broker.get_cluster_config(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_UPGRADE_PLAN)
def cluster_upgrade_plan(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster upgrade-plan operation.

    data validation handled in broker

    :return: List[Tuple(str, str)]
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])
    return vcd_broker.get_cluster_upgrade_plan(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_UPGRADE)
def cluster_upgrade(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster upgrade operation.

    data validation handled in broker

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])
    return vcd_broker.upgrade_cluster(data=data)


def _retain_cluster_list_common_properties(cluster_list: list, properties_to_retain: list = None) -> list:  # noqa: E501
    if properties_to_retain is None:
        properties_to_retain = []
    result = []
    for cluster_info in cluster_list:
        filtered_cluster_info = \
            {k: cluster_info.get(k) for k in properties_to_retain}
        result.append(filtered_cluster_info)
    return result


# TODO: Record telemetry in a different telemetry handler
@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_LIST)
def native_cluster_list(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: List
    """
    vcd_broker = VcdBroker(op_ctx)

    base_url = request_data['url']
    query_params = request_data.get(RequestKey.QUERY_PARAMS, {})
    page_number = int(query_params.get(PaginationKey.PAGE_NUMBER, CSE_PAGINATION_FIRST_PAGE_NUMBER))  # noqa: E501
    page_size = int(query_params.get(PaginationKey.PAGE_SIZE, CSE_PAGINATION_DEFAULT_PAGE_SIZE))  # noqa: E501 =
    query_params_others = {}
    for k, v in query_params.items():
        if k not in [PaginationKey.PAGE_NUMBER, PaginationKey.PAGE_SIZE]:
            query_params_others[k] = v

    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])

    vcd_clusters_info = vcd_broker.get_clusters_by_page(
        data=data, page_number=page_number, page_size=page_size)

    properties_to_retain = [
        'name',
        'vdc',
        'status',
        'org_name',
        'k8s_version',
        K8S_PROVIDER_KEY
    ]
    clusters = vcd_clusters_info[PaginationKey.VALUES]
    result = _retain_cluster_list_common_properties(clusters,
                                                    properties_to_retain)

    return server_utils.create_links_and_construct_paginated_result(
        base_url,
        result,
        vcd_clusters_info[PaginationKey.RESULT_TOTAL],
        page_number=page_number,
        page_size=page_size,
        query_params=query_params_others)


@record_user_action_telemetry(cse_operation=CseOperation.CLUSTER_LIST)
def cluster_list(request_data, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: List
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])
    vcd_clusters_info = vcd_broker.list_clusters(data=data)

    properties_to_retain = [
        'name',
        'vdc',
        'status',
        'org_name',
        'k8s_version',
        K8S_PROVIDER_KEY
    ]

    return _retain_cluster_list_common_properties(vcd_clusters_info,
                                                  properties_to_retain)


@record_user_action_telemetry(cse_operation=CseOperation.NODE_CREATE)
def node_create(request_data, op_ctx: ctx.OperationContext):
    """Request handler for node create operation.

    Required data: cluster_name, network_name
    Optional data and default values: org_name=None, ovdc_name=None,
        num_nodes=1, num_cpu=None, mb_memory=None, storage_profile_name=None,
        template_name=default, template_revision=default,
        ssh_key=None, rollback=True, enable_nfs=False,

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])
    return vcd_broker.create_nodes(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.NODE_DELETE)
def node_delete(request_data, op_ctx: ctx.OperationContext):
    """Request handler for node delete operation.

    Required data: cluster_name, node_names_list
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.INPUT_SPEC])
    return vcd_broker.delete_nodes(data=data)


@record_user_action_telemetry(cse_operation=CseOperation.NODE_INFO)
def node_info(request_data, op_ctx: ctx.OperationContext):
    """Request handler for node info operation.

    Required data: cluster_name, node_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    vcd_broker = VcdBroker(op_ctx)
    data = req_utils.flatten_request_data(
        request_data, [RequestKey.QUERY_PARAMS])
    return vcd_broker.get_node_info(data=data)



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import json
import os

import pyvcloud.vcd.client as vcd_client
import requests

import container_service_extension.client.constants as cli_constants
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
from container_service_extension.common.utils.core_utils import str_to_bool
from container_service_extension.exception.exceptions import CseResponseError
from container_service_extension.exception.minor_error_codes import MinorErrorCode  # noqa: E501
from container_service_extension.logging.logger import CLIENT_WIRE_LOGGER
from container_service_extension.logging.logger import NULL_LOGGER

wire_logger = NULL_LOGGER
if str_to_bool(os.getenv(cli_constants.ENV_CSE_CLIENT_WIRE_LOGGING)):
    wire_logger = CLIENT_WIRE_LOGGER


def _deserialize_response_content(response):
    """Convert utf-8 encoded string to a dict.

    Since the response is encoded in utf-8, it gets decoded to regular python
    string that will be a json string. That gets converted to python
    dictionary.

    Note: Do not use this method to process non-json response.content

    :param requests.models.Response response: object that includes attributes
        status code and content

    :return: response content as decoded dictionary

    :rtype: dict
    """
    if response.content:
        decoded = response.content.decode("utf-8")
        if len(decoded) > 0:
            return json.loads(decoded)
    return {}


def _response_to_exception(response):
    """Return exception object with appropriate messages.

    :param requests.models.Response response: object that has attributes
        status code and content

    :return: exception with proper error message

    :rtype: CseResponseError
    """
    error_message = ""
    minor_error_code = MinorErrorCode.DEFAULT_ERROR_CODE

    # in case of 401, content type is not set in the response and body is empty
    if response.status_code == requests.codes.unauthorized:
        error_message = 'Session has expired or user not logged in.' \
                        ' Please re-login.'
    elif response.status_code == requests.codes.too_many_requests:
        error_message = 'Server is busy. Please try again later.'
    elif 'json' in response.headers['Content-Type']:
        response_dict = _deserialize_response_content(response)
        if shared_constants.RESPONSE_MESSAGE_KEY in response_dict:
            message = response_dict[shared_constants.RESPONSE_MESSAGE_KEY]
            if isinstance(message, dict):
                if shared_constants.ERROR_DESCRIPTION_KEY in message:
                    error_message = message[shared_constants.ERROR_DESCRIPTION_KEY]  # noqa: E501
                if shared_constants.ERROR_MINOR_CODE_KEY in message:
                    minor_error_code = MinorErrorCode(message[shared_constants.ERROR_MINOR_CODE_KEY])  # noqa: E501
            else:
                error_message = message
    if not error_message:
        error_message = shared_constants.UNKNOWN_ERROR_MESSAGE

    raise CseResponseError(
        response.status_code, error_message, minor_error_code)


class CseClient:
    def __init__(self, client: vcd_client.Client):
        self._client = client
        self._uri = self._client.get_api_uri()
        self._request_page_size = 10

    def do_request(self,
                   uri: str,
                   method: str,
                   params: dict = None,
                   accept_type: str = None,
                   media_type: str = None,
                   payload: dict = None,
                   timeout: float = None):
        """."""
        wire_logger.debug(f"Made request to: {method} {uri}")
        wire_logger.debug(f"Accept type : {accept_type}")
        if params:
            wire_logger.debug(f"Query params : {params}")
        if payload:
            wire_logger.debug(f"Content-Type: {media_type}")
            wire_logger.debug(f"Content : {payload}")
        if timeout:
            wire_logger.debug(f"Timeout : {timeout}")

        # ToDo: Add support for timeout later
        response = self._client._do_request_prim(
            method,
            uri,
            self._client._session,
            accept_type=accept_type,
            contents=payload,
            media_type=media_type,
            params=params)

        wire_logger.debug(f"Request headers: {response.request.headers}")

        return response

    def iterate_results(self, base_url, filters=None):
        """Iterate over paginated response until all the results are obtained.

        :param str base_url: initial url
        :param dict filters: filters for making the API call
        :returns: Generator which yields values and a boolean indicating if
            more results are present
        :rtype: Generator[(List[dict], bool), None, None]

        page and pageSize is not needed if iteration should start from the
        first page.

        Example: Iterating through all the pages for get cluster
            (/cse/3.0/clusters) response from CSE server:
                self.iterate_results("https://vcd-api/api/cse/3.0/clusters")

        Example: Iterating through all the pages for get cluster starting from
            page 3:
                self.iterate_results("https://vcd-ip/api/cse/3.0/clusters?page=3&pageSize=25")
        """  # noqa: E501
        if filters is None:
            filters = {}
        # NOTE: This method is added here to reduce dependency between
        # cse_client package and the rest of the code.
        url = base_url
        while url:
            response = self.do_request(
                uri=url,
                method=shared_constants.RequestMethod.GET,
                params=filters,
                accept_type='application/json')
            processed_response = self.process_response(response)
            url = processed_response.get(shared_constants.PaginationKey.NEXT_PAGE_URI)  # noqa: E501
            yield processed_response[shared_constants.PaginationKey.VALUES], bool(url)  # noqa: E501

    @staticmethod
    def process_response(response):
        """Process the given response dictionary with following keys.

        Log the response if wire logging is enabled.

        If the value of status code is 2xx, return the response content, else
        raise exception with proper error message

        :param requests.models.Response response: object with attributes viz.
            status code and content
            status_code: http status code
            content: response result as string

        :return: decoded response content, if status code is 2xx or 429.

        :rtype: dict

        :raises CseResponseError: if response http status code is not 2xx or
            429
        """
        wire_logger.debug(f"Response status code: {response.status_code}")
        wire_logger.debug(f"Response headers: {response.headers}")

        response_content = _deserialize_response_content(response)
        wire_logger.debug(f"Response content: {response_content}")

        if response.status_code in [
            requests.codes.ok,
            requests.codes.created,
            requests.codes.accepted,
            requests.codes.no_content
        ]:
            return response_content

        raise _response_to_exception(response)



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools
import traceback

import requests

from container_service_extension.common.constants.shared_constants import ERROR_DESCRIPTION_KEY  # noqa: E501
from container_service_extension.common.constants.shared_constants import ERROR_MINOR_CODE_KEY   # noqa: E501
from container_service_extension.common.constants.shared_constants import RESPONSE_MESSAGE_KEY   # noqa: E501
from container_service_extension.exception.exceptions import CseRequestError
from container_service_extension.exception.minor_error_codes import MinorErrorCode # noqa: E501
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER


def handle_exception(func):
    """Decorate to trap exceptions and process them.

    If there are any exceptions, a dictionary containing the status code, body
        and stacktrace will be returned.

    This decorator should be applied only on those functions that constructs
    the final HTTP responses and also needs exception handler as additional
    behaviour.

    :param method func: decorated function

    :return: reference to the function that executes the decorated function
        and traps exceptions raised by it.
    """
    @functools.wraps(func)
    def exception_handler_wrapper(*args, **kwargs):
        result = {}
        try:
            result = func(*args, **kwargs)
        except Exception as err:
            if isinstance(err, CseRequestError):
                result['status_code'] = err.status_code
                minor_error_code = err.minor_error_code
            else:
                result['status_code'] = requests.codes.internal_server_error
                minor_error_code = MinorErrorCode.DEFAULT_ERROR_CODE

            error_string = str(err if err else '')
            result['body'] = {
                RESPONSE_MESSAGE_KEY: {
                    ERROR_MINOR_CODE_KEY: int(minor_error_code),
                    ERROR_DESCRIPTION_KEY: error_string
                }
            }
            LOGGER.error(traceback.format_exc())
        return result
    return exception_handler_wrapper



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from requests.exceptions import HTTPError

from container_service_extension.lib.nsxt.constants import RequestMethodVerb


class NSGroupManager(object):
    """Facilitate Create, Retrieve, Delete operations on NSGroups."""

    def __init__(self, nsxt_client):
        """Initialize a NSXGroupMnaager object.

        :param NSXTCLient nsxt_client: client to make NSX-T REST requests.
        """
        self._nsxt_client = nsxt_client

    def list_nsgroups(self):
        """List all NSGroups.

        :return: All NSGroups in the system as a list of dictionaries, where
            each dictionary represent a NSGroup.

        :rtype: list
        """
        resource_url_fragment = "ns-groups"
        response = self._nsxt_client.do_request(
            method=RequestMethodVerb.GET,
            resource_url_fragment=resource_url_fragment)

        nsgroups = response['results']
        return nsgroups

    def get_nsgroup(self, name=None, id=None):
        """Get information of a NSGroup identified by id or name.

        Identification by id takes precedence. Will return None if no matching
        NSGroup is found.

        :param str name: name of the NSGroup whose details are to be retrieved.
        :param str id: id of the NSGroup whose details are to be retrieved.

        :return: details of the NSGroup as a dictionary.

        :rtype: dict

        :raises HTTPError: If the underlying REST call fails with any status
            code other than 404.
        """
        if not id and not name:
            return

        if id:
            resource_url_fragment = f"ns-groups/{id}"
            try:
                response = self._nsxt_client.do_request(
                    method=RequestMethodVerb.GET,
                    resource_url_fragment=resource_url_fragment)
                return response
            except HTTPError as err:
                if err.response.status_code != 404:
                    raise
                else:
                    return

        nsgroups = self.list_nsgroups()
        for nsgroup in nsgroups:
            if nsgroup['display_name'].lower() == name.lower():
                return nsgroup

    def create_nsgroup(self, name, members=None, membership_criteria=None):
        """Create a new NSGroup.

        :param str name: name of the NSGroup to be created.
        :param list members: list of dictionaries. Where each dictionary
            represents a direct member of the NSGroup, generally expressed as
            NSGroupSimpleExpression JSON object.
        :param dict membership_criteria: dictionary representing indirect
            (evaluated) members of the NSGroup. Generally expressed as a
            NSGroupComplexExpression, NSGroupTagExpression or
            NSGroupSimpleExpression JSON object.

        :return: details of the newly created NSGroup as a dictionary.

        :rtype: dict
        """
        resource_url_fragment = "ns-groups"

        data = {}
        data['display_name'] = name
        if members:
            data['members'] = members
        if membership_criteria:
            data['membership_criteria'] = membership_criteria

        nodes_nsgroup = self._nsxt_client.do_request(
            method=RequestMethodVerb.POST,
            resource_url_fragment=resource_url_fragment,
            payload=data)

        return nodes_nsgroup

    def create_nsgroup_from_ipsets(self, name, ipset_ids):
        """Create a new NSGroup from a list of IPSets.

        :param str name: name of the NSGroup to be created.
        :param list ip_sets: list of ids (str) of IPSets, which are direct
            members of the NSGroup.

        :return: details of the newly created NSGroup as a dictionary.

        :rtype: dict
        """
        members = []
        for ipset_id in ipset_ids:
            member = {}
            member['resource_type'] = "NSGroupSimpleExpression"
            member['target_type'] = "IPSet"
            member['target_property'] = "id"
            member['op'] = "EQUALS"
            member['value'] = ipset_id

            members.append(member)

        return self.create_nsgroup(name, members=members)

    def delete_nsgroup(self, name=None, id=None, force=False):
        """Delete a NSGroup identified by id or name.

        Identification by id takes precedence. Will return False if no matching
        NSGroup is found.

        :param str name: name of the NSGroup to be deleted..
        :param str id: id of the NSGroup to be deleted.
        :param bool force: if True, a force deletion will be attempted.

        :return: True, if the delete operation is successful, else False.

        :rtype: bool
        """
        if not name and not id:
            return False

        if not id:
            nsgroup = self.get_nsgroup(name)
            if nsgroup:
                id = nsgroup['id']
            else:
                self._nsxt_client.LOGGER.debug(f"NSGroup : {name} not found. "
                                               "Unable to delete.")
                return False
        resource_url_fragment = f"ns-groups/{id}"
        if force:
            resource_url_fragment += "?force=true"

        self._nsxt_client.do_request(
            method=RequestMethodVerb.DELETE,
            resource_url_fragment=resource_url_fragment)
        return True



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
import semantic_version

from container_service_extension.common.constants.server_constants import FlattenedClusterSpecKey1X  # noqa: E501
from container_service_extension.common.constants.server_constants import ThreadLocalData  # noqa: E501
from container_service_extension.common.constants.shared_constants import ClusterAclKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_DEFAULT_PAGE_SIZE  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_FIRST_PAGE_NUMBER  # noqa: E501
from container_service_extension.common.constants.shared_constants import PaginationKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
import container_service_extension.common.thread_local_data as thread_local_data  # noqa: E501
import container_service_extension.common.utils.server_utils as server_utils
import container_service_extension.exception.exceptions as exceptions
import container_service_extension.lib.telemetry.constants as telemetry_constants  # noqa: E501
import container_service_extension.lib.telemetry.telemetry_handler as telemetry_handler  # noqa: E501
import container_service_extension.rde.backend.cluster_service_factory as cluster_service_factory  # noqa: E501
import container_service_extension.rde.constants as rde_constants
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
import container_service_extension.rde.models.common_models as common_models
import container_service_extension.rde.models.rde_factory as rde_factory
import container_service_extension.rde.utils as rde_utils
import container_service_extension.security.context.operation_context as ctx
import container_service_extension.server.request_handlers.cluster_handler as cluster_handler  # noqa: E501
import container_service_extension.server.request_handlers.request_utils as request_utils  # noqa: E501

_OPERATION_KEY = 'operation'
_SUPPORTED_API_VERSION = '35.0'


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_APPLY)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_create(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster create operation.

    :return: Defined entity of the native cluster
    :rtype: container_service_extension.def_.models.DefEntity
    """
    rde_in_use = server_utils.get_rde_version_in_use()
    input_entity: dict = data[RequestKey.INPUT_SPEC]
    payload_version = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION_RDE_1_0.value)  # noqa: E501
    # Reject request >= 2.0
    _raise_error_if_unsupported_payload_version(payload_version)
    # Convert the input entity to runtime rde format
    converted_native_entity: AbstractNativeEntity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        rde_data: dict = cluster_handler.cluster_create(
            data={RequestKey.INPUT_SPEC: converted_native_entity.to_dict()},
            op_ctx=op_ctx)
        new_rde = common_models.DefEntity(**rde_data)
    else:
        # Based on the runtime rde, call the appropriate backend method.
        svc = cluster_service_factory.ClusterServiceFactory(op_ctx).get_cluster_service()  # noqa: E501
        new_rde: common_models.DefEntity = svc.create_cluster(converted_native_entity)  # noqa: E501

    # convert the created rde back to input rde version
    return _convert_rde_to_1_0_format(new_rde.to_dict())


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_APPLY)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_resize(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster resize operation.

    Validate data before actual resize is delegated to cluster service.

    :return: Defined entity of the native cluster
    :rtype: container_service_extension.def_.models.DefEntity
    """
    rde_in_use = server_utils.get_rde_version_in_use()
    input_entity: dict = data[RequestKey.INPUT_SPEC]
    payload_version = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION_RDE_1_0.value)  # noqa: E501
    # Reject request >= 2.0
    _raise_error_if_unsupported_payload_version(payload_version)
    # Convert the input entity to runtime rde format
    converted_native_entity: AbstractNativeEntity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        data[RequestKey.INPUT_SPEC] = converted_native_entity.to_dict()
        rde_data: dict = cluster_handler.cluster_update(data=data, op_ctx=op_ctx)  # noqa: E501
        new_rde = common_models.DefEntity(**rde_data)
    else:
        # Based on the runtime rde, call the appropriate backend method.
        svc = cluster_service_factory.ClusterServiceFactory(op_ctx).get_cluster_service()  # noqa: E501
        cluster_id = data[RequestKey.CLUSTER_ID]
        curr_entity = svc.entity_svc.get_entity(cluster_id)
        request_utils.validate_request_payload(
            converted_native_entity.spec.to_dict(), curr_entity.entity.spec.to_dict(),  # noqa: E501
            exclude_fields=[FlattenedClusterSpecKey1X.WORKERS_COUNT.value,
                            FlattenedClusterSpecKey1X.NFS_COUNT.value,
                            FlattenedClusterSpecKey1X.EXPOSE.value])
        new_rde: common_models.DefEntity = svc.resize_cluster(cluster_id, converted_native_entity)  # noqa: E501

    # convert the resized rde back to input rde version
    return _convert_rde_to_1_0_format(new_rde.to_dict())


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_DELETE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_delete(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster delete operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    rde_in_use = server_utils.get_rde_version_in_use()
    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        rde_data = cluster_handler.cluster_delete(data=data, op_ctx=op_ctx)
        new_rde = common_models.DefEntity(**rde_data)
    else:
        svc = cluster_service_factory.ClusterServiceFactory(op_ctx).get_cluster_service(rde_in_use)  # noqa: E501
        cluster_id = data[RequestKey.CLUSTER_ID]
        new_rde = svc.delete_cluster(cluster_id)

    # convert the deleted rde data back to input rde version
    return _convert_rde_to_1_0_format(new_rde.to_dict())


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_INFO)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_info(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster info operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        rde_data: dict = cluster_handler.cluster_info(data=data, op_ctx=op_ctx)
        new_rde = common_models.DefEntity(**rde_data)
    else:
        svc = cluster_service_factory.ClusterServiceFactory(op_ctx).get_cluster_service(rde_in_use)  # noqa: E501
        cluster_id = data[RequestKey.CLUSTER_ID]
        new_rde = svc.get_cluster_info(cluster_id)

    return _convert_rde_to_1_0_format(new_rde.to_dict())


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_CONFIG)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_config(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster config operation.

    Required data: cluster_id

    :return: Dict
    """
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        return cluster_handler.cluster_config(data=data, op_ctx=op_ctx)

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)
    cluster_id = data[RequestKey.CLUSTER_ID]
    return svc.get_cluster_config(cluster_id)


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_UPGRADE_PLAN)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_upgrade_plan(data, op_ctx: ctx.OperationContext):
    """Request handler for cluster upgrade-plan operation.

    :return: List[Tuple(str, str)]
    """
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        return cluster_handler.cluster_upgrade_plan(data=data, op_ctx=op_ctx)

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)
    return svc.get_cluster_upgrade_plan(data[RequestKey.CLUSTER_ID])


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_UPGRADE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_upgrade(data, op_ctx: ctx.OperationContext):
    """Request handler for cluster upgrade operation.

    Validate data before actual upgrade is delegated to cluster service.

    :return: Dict
    """
    rde_in_use = server_utils.get_rde_version_in_use()
    input_entity: dict = data[RequestKey.INPUT_SPEC]
    payload_version = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION_RDE_1_0.value)  # noqa: E501
    # Reject request >= 2.0
    _raise_error_if_unsupported_payload_version(payload_version)
    # Convert the input entity to runtime rde format
    converted_native_entity: AbstractNativeEntity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        data[RequestKey.INPUT_SPEC] = converted_native_entity.to_dict()
        rde_data: dict = cluster_handler.cluster_update(data=data, op_ctx=op_ctx)  # noqa: E501
        new_rde = common_models.DefEntity(**rde_data)
    else:
        # Based on the runtime rde, call the appropriate backend method.
        svc = cluster_service_factory.ClusterServiceFactory(op_ctx).get_cluster_service()  # noqa: E501
        cluster_id = data[RequestKey.CLUSTER_ID]
        curr_entity = svc.entity_svc.get_entity(cluster_id)
        request_utils.validate_request_payload(
            converted_native_entity.spec.to_dict(), curr_entity.entity.spec.to_dict(),  # noqa: E501
            exclude_fields=[FlattenedClusterSpecKey1X.TEMPLATE_NAME.value,
                            FlattenedClusterSpecKey1X.TEMPLATE_REVISION.value])
        new_rde: common_models.DefEntity = svc.upgrade_cluster(cluster_id, converted_native_entity)  # noqa: E501

    # convert the upgraded rde back to input rde version
    return _convert_rde_to_1_0_format(new_rde.to_dict())


# TODO: Record telemetry in a different telemetry handler
@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_UPGRADE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def native_cluster_list(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    :return: List
    """
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        response_data: dict = cluster_handler.native_cluster_list(data=data, op_ctx=op_ctx)  # noqa: E501
        rde_list: list[dict] = response_data[PaginationKey.VALUES]
        formatted_rde_list = [_convert_rde_to_1_0_format(rde_data) for rde_data in rde_list]  # noqa: E501
        response_data[PaginationKey.VALUES] = formatted_rde_list
        return response_data

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)
    filters = data.get(RequestKey.QUERY_PARAMS, {})
    page_number = int(filters.get(PaginationKey.PAGE_NUMBER,
                                  CSE_PAGINATION_FIRST_PAGE_NUMBER))
    page_size = int(filters.get(PaginationKey.PAGE_SIZE,
                                CSE_PAGINATION_DEFAULT_PAGE_SIZE))

    # remove page number and page size from the filters as it is treated
    # differently from other filters
    if PaginationKey.PAGE_NUMBER in filters:
        del filters[PaginationKey.PAGE_NUMBER]
    if PaginationKey.PAGE_SIZE in filters:
        del filters[PaginationKey.PAGE_SIZE]

    # response needs to paginated
    result = svc.get_clusters_by_page(filters=filters)
    clusters = [def_entity.to_dict() for def_entity in result[PaginationKey.VALUES]]  # noqa: E501

    uri = data['url']
    return server_utils.create_links_and_construct_paginated_result(
        uri,
        clusters,
        result_total=result[PaginationKey.RESULT_TOTAL],
        page_number=page_number,
        page_size=page_size,
        query_params=filters)


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_ACL_LIST)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_acl_info(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster acl list operation."""
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        return cluster_handler.cluster_acl_info(data=data, op_ctx=op_ctx)

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)
    cluster_id = data[RequestKey.CLUSTER_ID]
    query = data.get(RequestKey.QUERY_PARAMS, {})
    page = int(query.get(PaginationKey.PAGE_NUMBER, CSE_PAGINATION_FIRST_PAGE_NUMBER))  # noqa: E501
    page_size = int(query.get(PaginationKey.PAGE_SIZE, CSE_PAGINATION_DEFAULT_PAGE_SIZE))  # noqa: E501
    result: dict = svc.get_cluster_acl_info(cluster_id, page, page_size)

    uri = data['url']
    return server_utils.create_links_and_construct_paginated_result(
        base_uri=uri,
        values=result.get(PaginationKey.VALUES, []),
        result_total=result.get(PaginationKey.RESULT_TOTAL, 0),
        page_number=page,
        page_size=page_size)


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_ACL_UPDATE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_acl_update(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster acl update operation."""
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        return cluster_handler.cluster_acl_update(data=data, op_ctx=op_ctx)

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)
    cluster_id = data[RequestKey.CLUSTER_ID]
    update_acl_entries = data.get(RequestKey.INPUT_SPEC, {}).get(ClusterAclKey.ACCESS_SETTING)  # noqa: E501
    svc.update_cluster_acl(cluster_id, update_acl_entries)


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_CLUSTER_LIST)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_list(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    :return: List
    """
    rde_in_use = server_utils.get_rde_version_in_use()

    # Redirect to generic handler if the backend supports RDE-2.0
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        rde_list: list[dict] = cluster_handler.cluster_list(data=data, op_ctx=op_ctx)  # noqa: E501
        return [_convert_rde_to_1_0_format(rde_data) for rde_data in rde_list]

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)

    # response should not be paginated
    return [def_entity.to_dict() for def_entity in
            svc.list_clusters(data.get(RequestKey.QUERY_PARAMS, {}))]


@request_utils.cluster_api_exception_handler
def node_create(request_data, op_ctx: ctx.OperationContext):
    """Request handler for node create operation.

    Required data: cluster_name, network_name
    Optional data and default values: org_name=None, ovdc_name=None,
        num_nodes=1, num_cpu=None, mb_memory=None, storage_profile_name=None,
        template_name=default, template_revision=default,
        ssh_key=None, rollback=True, enable_nfs=False,

    (data validation handled in broker)

    :return: Dict
    """
    raise NotImplementedError


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V35_NODE_DELETE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def nfs_node_delete(data, op_ctx: ctx.OperationContext):
    """Request handler for node delete operation.

    Required data: cluster_name, node_names_list
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    rde_in_use = server_utils.get_rde_version_in_use()
    if semantic_version.Version(rde_in_use) >= semantic_version.Version(rde_constants.RDEVersion.RDE_2_0_0.value):  # noqa: E501
        cluster_dict = cluster_handler.nfs_node_delete(data=data, op_ctx=op_ctx)  # noqa: E501
        # NOTE: cluster_handler is always expected to return RDE version
        #   2.0.0 or more. Hence a convertion to 1.0 is needed
        rde_class = rde_factory.get_rde_model(rde_in_use)
        cluster_entity: AbstractNativeEntity = rde_class.from_dict(cluster_dict['entity'])  # noqa: E501
        new_native_entity: AbstractNativeEntity = rde_utils.convert_runtime_rde_to_input_rde_version_format(  # noqa: E501
            cluster_entity, rde_constants.RDEVersion.RDE_1_0_0)
        cluster_dict['entity'] = new_native_entity.to_dict()
        cluster_dict['entityType'] = common_models.EntityType.NATIVE_ENTITY_TYPE_1_0_0.value.id  # noqa: E501
        return cluster_dict

    svc = cluster_service_factory.ClusterServiceFactory(op_ctx). \
        get_cluster_service(rde_in_use)
    cluster_id = data[RequestKey.CLUSTER_ID]
    node_name = data[RequestKey.NODE_NAME]

    telemetry_handler.record_user_action_details(
        cse_operation=telemetry_constants.CseOperation.V35_NODE_DELETE,
        cse_params={
            telemetry_constants.PayloadKey.CLUSTER_ID: cluster_id,
            telemetry_constants.PayloadKey.NODE_NAME: node_name,
            telemetry_constants.PayloadKey.SOURCE_DESCRIPTION: thread_local_data.get_thread_local_data(ThreadLocalData.USER_AGENT)   # noqa: E501
        }
    )

    return svc.delete_nodes(cluster_id, [node_name]).to_dict()


@request_utils.cluster_api_exception_handler
def node_info(request_data, op_ctx: ctx.OperationContext):
    """Request handler for node info operation.

    Required data: cluster_name, node_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    raise NotImplementedError


def _raise_error_if_unsupported_payload_version(payload_version: str):
    input_rde_version = rde_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION.get(payload_version, rde_constants.RDEVersion.RDE_1_0_0.value)  # noqa: E501
    supported_rde_version = rde_utils.get_rde_version_introduced_at_api_version(_SUPPORTED_API_VERSION)  # noqa: E501
    if semantic_version.Version(input_rde_version) > semantic_version.Version(supported_rde_version):  # noqa: E501
        raise exceptions.BadRequestError(f"Unsupported payload version: {payload_version}")  # noqa: E501


def _convert_rde_to_1_0_format(rde_data: dict) -> dict:
    """Convert defined entity to RDE 1.0.

    :param DefEntity rde_data: Defined entity dictionary
    :return: converted defined entity
    :rtype: dict
    """
    new_rde = common_models.DefEntity(**rde_data)
    new_native_entity: AbstractNativeEntity = rde_utils.convert_runtime_rde_to_input_rde_version_format(  # noqa: E501
        new_rde.entity, rde_constants.RDEVersion.RDE_1_0_0)
    new_rde.entity = new_native_entity
    new_rde.entityType = common_models.EntityType.NATIVE_ENTITY_TYPE_1_0_0.value.get_id()  # noqa: E501
    return new_rde.to_dict()



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Thread utils for managing multiple threads in the server."""

import functools
import threading

import container_service_extension.common.thread_local_data as thread_local_data  # noqa: E501


def _transfer_thread_local_data_wrapper(func):
    cur_thread_data = thread_local_data.get_thread_local_data_as_dict()

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        thread_local_data.set_thread_local_data_from_dict(cur_thread_data)
        func(*args, **kwargs)
        thread_local_data.reset_thread_local_data()
    return wrapper


def run_async(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        thread_local_data_wrapper = _transfer_thread_local_data_wrapper(func)
        t = threading.Thread(name=generate_thread_name(func.__name__),
                             target=thread_local_data_wrapper, args=args, kwargs=kwargs,  # noqa: E501
                             daemon=True)
        t.start()
        return t
    return wrapper


def generate_thread_name(function_name):
    parent_thread_id = threading.current_thread().ident
    return function_name + ':' + str(parent_thread_id)



# coding: utf-8

"""
    TKG Kubernetes API

    This API provides to vCD tenants the means to provision (create and update) Tanzu Kubernetes Grid clusters. This is complementary to the defined-entity APIs:    GET /cloudapi/1.0.0/entities/urn:vcloud:entity:vmware.tkgcluster:1.0.0:{id} which allows to retrieve the clusters created by the API presented here. This is why you will not find here a GET operation for the corresponding entity.   # noqa: E501

    OpenAPI spec version: 1.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class TkgClusterSpecDistribution(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'version': 'str'
    }

    attribute_map = {
        'version': 'version'
    }

    def __init__(self, version=None):  # noqa: E501
        """TkgClusterSpecDistribution - a model defined in Swagger"""  # noqa: E501

        self._version = None
        self.discriminator = None

        self.version = version

    @property
    def version(self):
        """Gets the version of this TkgClusterSpecDistribution.  # noqa: E501

        Specifies the software version of the Kubernetes distribution to install on cluster nodes using semantic version notation. Can specify the fully qualified version or use version shortcuts, such as \"version: v1.16.6\", which is resolved to the most recent image matching that patch version, or \"version: v1.16\", which is resolved to the most recent matching patch version. The resolved version displays as the \"fullVersion\" on the cluster description after you have created it.   # noqa: E501

        :return: The version of this TkgClusterSpecDistribution.  # noqa: E501
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this TkgClusterSpecDistribution.

        Specifies the software version of the Kubernetes distribution to install on cluster nodes using semantic version notation. Can specify the fully qualified version or use version shortcuts, such as \"version: v1.16.6\", which is resolved to the most recent image matching that patch version, or \"version: v1.16\", which is resolved to the most recent matching patch version. The resolved version displays as the \"fullVersion\" on the cluster description after you have created it.   # noqa: E501

        :param version: The version of this TkgClusterSpecDistribution.  # noqa: E501
        :type: str
        """
        if version is None:
            raise ValueError("Invalid value for `version`, must not be `None`")  # noqa: E501

        self._version = version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TkgClusterSpecDistribution):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
from dataclasses import dataclass
from enum import Enum
from enum import unique
from typing import List, Optional

from dataclasses_json import dataclass_json, Undefined

from container_service_extension.common.constants import shared_constants as shared_constants  # noqa: E501
import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
from container_service_extension.rde import utils as def_utils
from container_service_extension.rde.behaviors.behavior_model import BehaviorAcl, BehaviorOperation  # noqa: E501
from container_service_extension.rde.constants import \
    DEF_ENTITY_TYPE_ID_PREFIX, DEF_INTERFACE_ID_PREFIX, Nss, RDEMetadataKey, \
    RDEVersion, RuntimeRDEVersion, SchemaFile, Vendor
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
from container_service_extension.rde.models.rde_factory import get_rde_model
from container_service_extension.rde.utils import load_rde_schema


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass(frozen=True)
class DefInterface:
    """Provides interface for the defined entity type."""

    name: str
    id: Optional[str] = None
    readonly: bool = False
    vendor: str = Vendor.CSE.value
    nss: str = Nss.KUBERNETES.value
    version: str = '1.0.0'

    def get_id(self):
        """Get or generate interface id.

        Example: urn:vcloud:interface:cse.native:1.0.0.

        By no means, id generation in this method, guarantees the actual
        interface registration with vCD.
        """
        if self.id is None:
            return def_utils.\
                generate_interface_id(self.vendor, self.nss, self.version)
        else:
            return self.id


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass(frozen=True)
class DefEntityType:
    """Defined Entity type schema for the apiVersion = 35.0."""

    name: str
    description: str
    schema: dict
    interfaces: List[str]
    version: str
    id: Optional[str] = None
    externalId: Optional[str] = None
    readonly: bool = False
    vendor: str = Vendor.CSE.value
    nss: str = Nss.NATIVE_CLUSTER.value

    def get_id(self):
        """Get or generate entity type id.

        Example : "urn:vcloud:interface:cse.native:1.0.0

        By no means, id generation in this method, guarantees the actual
        entity type registration with vCD.
        """
        if self.id is None:
            return def_utils.\
                generate_entity_type_id(self.vendor, self.nss, self.version)
        else:
            return self.id


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass(frozen=True)
class DefEntityType2_0:
    """Defined Entity type schema for the apiVersion = 36.0."""

    name: str
    description: str
    schema: dict
    interfaces: List[str]
    version: str
    id: str = None
    externalId: Optional[str] = None
    readonly: bool = False
    vendor: str = Vendor.CSE.value
    nss: str = Nss.NATIVE_CLUSTER.value
    hooks: dict = None

    def get_id(self):
        """Get or generate entity type id.

        Example : "urn:vcloud:interface:cse.native:1.0.0

        By no means, id generation in this method, guarantees the actual
        entity type registration with vCD.
        """
        if self.id is None:
            return def_utils.\
                generate_entity_type_id(self.vendor, self.nss, self.version)
        else:
            return self.id


@dataclass_json
@dataclass
class Owner:
    name: Optional[str] = None
    id: Optional[str] = None


@dataclass_json
@dataclass
class Org:
    name: Optional[str] = None
    id: Optional[str] = None


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class DefEntity:
    """Represents defined entity instance.

    If dictionaries are passed as arguments, the constructor auto-converts
    them into the expected class instances.
    """

    name: str
    entity: AbstractNativeEntity
    id: Optional[str] = None
    entityType: Optional[str] = None
    externalId: Optional[str] = None
    state: Optional[str] = None
    owner: Optional[Owner] = None
    org: Optional[Org] = None

    def __init__(self, entity: AbstractNativeEntity, entityType: str,
                 name: str = None, id: str = None,
                 externalId: str = None, state: str = None,
                 owner: Owner = None, org: Org = None):
        self.id = id
        self.entityType = entityType

        # Get the entity type version from entity type urn
        entity_type_version = self.entityType.split(":")[-1]
        self.entity = entity
        if isinstance(entity, dict):
            # Parse the entity to the right entity class
            NativeEntityClass = get_rde_model(entity_type_version)
            self.entity = NativeEntityClass.from_dict(entity)

        self.name = name

        if not self.name:
            if entity_type_version == RDEVersion.RDE_1_0_0:
                self.name = self.entity.metadata.cluster_name
            else:
                self.name = self.entity.metadata.name

        self.externalId = externalId
        self.state = state
        self.owner = Owner(**owner) if isinstance(owner, dict) else owner
        self.org = Org(**org) if isinstance(org, dict) else org


@dataclass()
class Ovdc:
    k8s_runtime: List[str]
    ovdc_name: Optional[str] = None
    ovdc_id: Optional[str] = None
    org_name: Optional[str] = None
    remove_cp_from_vms_on_disable: bool = False


@dataclass()
class Tenant:
    name: Optional[str] = None
    id: Optional[str] = None


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass(frozen=False)
class ClusterAclEntry:
    accessLevelId: Optional[server_constants.AclAccessLevelId] = None
    memberId: Optional[str] = None
    id: Optional[str] = None
    grantType: Optional[str] = None
    objectId: Optional[str] = None
    username: Optional[str] = None
    tenant: Optional[Tenant] = None

    def construct_filtered_dict(self, include=None):
        if include is None:
            include = []
        orig_dict = self.to_dict()
        include_set = set(include)
        filtered_dict = {}
        for key, value in orig_dict.items():
            if key in include_set:
                filtered_dict[key] = value
        return filtered_dict


# NOTE: Only used for cluster list operation to get entities by interface
# include the following properties:
@dataclass()
class GenericClusterEntity:
    # Properties being used for List operation attributes representing them
    # for native and TKG clusters:
    # name:
    #   def_entity.name
    # vdc:
    #   def_entity.entity.metadata.ovdc_name for native
    #   de_entity.entity.metadata.virtualDataCenterName for TKG
    # org:
    #   def_entity.org.name for both native and TKG
    # runtime:
    #   def_entity.entity.kind for both native and TKG
    # version:
    #   def_entity.entity.status.kubernetes for native,
    #   def_entity.entity.spec.distribution.version for TKG
    # status:
    #   def_entity.entity.status.phase for both native and TKG
    # owner:
    #   def_entity.owner.name for both native and TKG
    name: Optional[str] = None
    org: Optional[Org] = None
    entity = None
    owner: Optional[Owner] = None

    def __init__(self, name: str, org: Org, entityType: str, entity,
                 owner: Owner, **kwargs):
        self.name = name
        self.org = Org(**org) if isinstance(org, dict) else org
        self.entityType = entityType
        entity_dict = entity.to_dict() if not isinstance(entity, dict) else entity  # noqa: E501
        if entity_dict['kind'] in \
                [shared_constants.ClusterEntityKind.NATIVE.value,
                 shared_constants.ClusterEntityKind.TKG_PLUS.value]:
            # Get the entity type version from entity type urn
            entity_type_version = self.entityType.split(":")[-1]
            # Parse the entity to the right entity class
            NativeEntityClass = get_rde_model(entity_type_version)
            self.entity: AbstractNativeEntity = \
                NativeEntityClass.from_dict(entity_dict) if isinstance(entity, dict) else entity  # noqa: E501
        elif entity_dict['kind'] == \
                shared_constants.ClusterEntityKind.TKG_S.value:
            self.entity = TKGEntity(**entity) if isinstance(entity, dict) else entity  # noqa: E501
        else:
            raise Exception("Invalid cluster kind")
        self.owner = Owner(**owner) if isinstance(owner, dict) else owner


@dataclass()
class TKGDistribution:
    version: Optional[str] = None

    def __init__(self, version: str, **kwargs):
        self.version = version


@dataclass()
class TKGSpec:
    distribution: Optional[TKGDistribution] = None

    def __init__(self, distribution: TKGDistribution, **kwargs):
        self.distribution = TKGDistribution(**distribution) \
            if isinstance(distribution, dict) else distribution


@dataclass()
class TKGStatus:
    phase: Optional[str] = None

    def __init__(self, phase: str, **kwargs):
        self.phase = phase


@dataclass()
class TKGMetadata:
    virtualDataCenterName: Optional[str] = None

    def __init__(self, virtualDataCenterName: str, **kwargs):
        self.virtualDataCenterName = virtualDataCenterName


@dataclass()
class TKGEntity:
    kind: Optional[str] = None
    spec: Optional[TKGSpec] = None
    status: Optional[TKGStatus] = None
    metadata: Optional[TKGMetadata] = None

    def __init__(self, kind: str, spec: TKGSpec, status: TKGStatus,
                 metadata: TKGMetadata, **kwargs):
        self.kind = kind
        self.spec = TKGSpec(**spec) if isinstance(spec, dict) else spec
        self.status = TKGStatus(**status) \
            if isinstance(status, dict) else status
        self.metadata = TKGMetadata(**metadata) \
            if isinstance(metadata, dict) else metadata


@unique
class K8Interface(Enum):
    VCD_INTERFACE = DefInterface(name='Kubernetes', vendor=Vendor.VMWARE.value,
                                 nss=Nss.KUBERNETES.value, version='1.0.0',
                                 id=f"{DEF_INTERFACE_ID_PREFIX}:{Vendor.VMWARE.value}:{Nss.KUBERNETES.value}:1.0.0")  # noqa: E501
    CSE_INTERFACE = DefInterface(name='CSE_K8s_interface', vendor=Vendor.CSE.value,  # noqa: E501
                                 nss=Nss.KUBERNETES.value, version='1.0.0',
                                 id=f"{DEF_INTERFACE_ID_PREFIX}:{Vendor.CSE.value}:{Nss.KUBERNETES.value}:1.0.0")  # noqa: E501


@unique
class EntityType(Enum):
    NATIVE_ENTITY_TYPE_1_0_0 = DefEntityType(name='nativeClusterEntityType',
                                             id=f"{DEF_ENTITY_TYPE_ID_PREFIX}:{Vendor.CSE.value}:{Nss.NATIVE_CLUSTER}:1.0.0",  # noqa: E501
                                             schema=load_rde_schema(SchemaFile.SCHEMA_1_0_0),  # noqa: E501
                                             interfaces=[K8Interface.VCD_INTERFACE.value.id],  # noqa: E501
                                             version='1.0.0',
                                             vendor=Vendor.CSE.value,
                                             nss=Nss.NATIVE_CLUSTER.value,
                                             description='')
    NATIVE_ENTITY_TYPE_2_0_0 = DefEntityType2_0(name='nativeClusterEntityType',
                                                id=f"{DEF_ENTITY_TYPE_ID_PREFIX}:{Vendor.CSE.value}:{Nss.NATIVE_CLUSTER}:2.0.0",  # noqa: E501
                                                schema=load_rde_schema(SchemaFile.SCHEMA_2_0_0),  # noqa: E501
                                                interfaces=[
                                                    K8Interface.VCD_INTERFACE.value.id,  # noqa: E501
                                                    K8Interface.CSE_INTERFACE.value.id],  # noqa: E501
                                                version='2.0.0',
                                                vendor=Vendor.CSE.value,
                                                nss=Nss.NATIVE_CLUSTER.value,
                                                description='',
                                                hooks={
                                                    'PostCreate': BehaviorOperation.CREATE_CLUSTER.value.id,  # noqa: E501
                                                    'PostUpdate': BehaviorOperation.UPDATE_CLUSTER.value.id,  # noqa: E501
                                                    'PreDelete': BehaviorOperation.DELETE_CLUSTER.value.id}  # noqa: E501
                                                )
    TKG_ENTITY_TYPE_1_0_0 = DefEntityType(name='TKG Cluster',
                                          id=f"{DEF_ENTITY_TYPE_ID_PREFIX}:{Vendor.VMWARE.value}:{Nss.TKG}:1.0.0",  # noqa: E501
                                          schema={},
                                          interfaces=[K8Interface.VCD_INTERFACE.value.id],  # noqa: E501
                                          version='1.0.0',
                                          vendor=Vendor.VMWARE.value,
                                          nss=Nss.TKG.value,
                                          description='')


# Key: Represents the Runtime RDE version used by CSE server for any given environment.  # noqa: E501
# Value: Details about all of the RDE constructs related to the specified RDE version.  # noqa: E501
MAP_RDE_VERSION_TO_ITS_METADATA = {

    RuntimeRDEVersion.RDE_1_X: {
        RDEMetadataKey.INTERFACES: [K8Interface.VCD_INTERFACE.value],
        RDEMetadataKey.ENTITY_TYPE: EntityType.NATIVE_ENTITY_TYPE_1_0_0.value,
    },

    RuntimeRDEVersion.RDE_2_X: {
        RDEMetadataKey.INTERFACES: [K8Interface.VCD_INTERFACE.value,
                                    K8Interface.CSE_INTERFACE.value],
        RDEMetadataKey.ENTITY_TYPE: EntityType.NATIVE_ENTITY_TYPE_2_0_0.value,

        RDEMetadataKey.INTERFACE_TO_BEHAVIORS_MAP: {
            K8Interface.CSE_INTERFACE.value.id:
                [BehaviorOperation.CREATE_CLUSTER.value,
                 BehaviorOperation.UPDATE_CLUSTER.value,
                 BehaviorOperation.DELETE_CLUSTER.value,
                 BehaviorOperation.DELETE_NFS_NODE.value]
        },
        RDEMetadataKey.ENTITY_TYPE_TO_OVERRIDABLE_BEHAVIORS_MAP: {
            EntityType.NATIVE_ENTITY_TYPE_2_0_0.value.id:
                [BehaviorOperation.GET_KUBE_CONFIG.value]
        },
        RDEMetadataKey.BEHAVIOR_TO_ACL_MAP: {
            EntityType.NATIVE_ENTITY_TYPE_2_0_0.value.id:
                [BehaviorAcl.CREATE_CLUSTER_ACL.value,
                 BehaviorAcl.UPDATE_CLUSTER_ACL.value,
                 BehaviorAcl.DELETE_CLUSTER_ACL.value,
                 BehaviorAcl.KUBE_CONFIG_ACL.value,
                 BehaviorAcl.DELETE_NFS_NODE_ACL.value]
        }
    }
}



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Claus

from container_service_extension.common.thread_local_data import init_thread_local_data  # noqa: E501
import container_service_extension.logging.logger
from container_service_extension.logging.logger import configure_null_logger

init_thread_local_data()
container_service_extension.logging.logger.configure_all_file_loggers()
configure_null_logger()
